import uuid
from datetime import datetime, timezone
from enum import Enum
from typing import Dict, List, Literal, Optional, Set

from pydantic import UUID4, BaseModel, Field


class RunType(Enum):
    SIMPLE = "simple"
    QUERY = "query"
    QUERY_BASE = "query_base"
    QUERY_DIFF = "query_diff"
    VALUE_DIFF = "value_diff"
    VALUE_DIFF_DETAIL = "value_diff_detail"
    SCHEMA_DIFF = "schema_diff"
    PROFILE = "profile"
    PROFILE_DIFF = "profile_diff"
    ROW_COUNT = "row_count"
    ROW_COUNT_DIFF = "row_count_diff"
    LINEAGE_DIFF = "lineage_diff"
    TOP_K_DIFF = "top_k_diff"
    HISTOGRAM_DIFF = "histogram_diff"

    def __str__(self):
        return self.value


class RunProgress(BaseModel):
    message: str
    percentage: float


class RunStatus(Enum):
    FINISHED = "finished"
    FAILED = "failed"
    CANCELLED = "cancelled"
    RUNNING = "running"
    # This is a special status only in v0.36.0. Replaced by FINISHED. To be removed in the future.
    SUCCESSFUL = "successful"


class Run(BaseModel):
    type: RunType
    name: Optional[str] = None
    params: Optional[dict] = None
    check_id: Optional[UUID4] = None
    result: Optional[dict] = None
    error: Optional[str] = None
    status: Optional[RunStatus] = None
    progress: Optional[RunProgress] = None
    run_id: UUID4 = Field(default_factory=uuid.uuid4)
    run_at: str = Field(default_factory=lambda: datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ"))


class Check(BaseModel):
    name: str
    description: Optional[str] = None
    type: RunType
    params: Optional[dict] = {}
    view_options: Optional[dict] = {}
    check_id: UUID4 = Field(default_factory=uuid.uuid4)
    is_checked: bool = False
    is_preset: bool = False
    created_at: datetime = Field(default_factory=lambda: datetime.now(timezone.utc).replace(microsecond=0))
    updated_at: datetime = Field(default_factory=lambda: datetime.now(timezone.utc).replace(microsecond=0))

    def merge(self, other) -> bool:
        if not self.is_preset and not other.is_preset and self.check_id != other.check_id:
            raise ValueError(f"check_id mismatch: {self.check_id} != {other.check_id}")

        if self.updated_at and other.updated_at and self.updated_at >= other.updated_at:
            return False

        self.name = other.name
        self.description = other.description
        self.type = other.type
        self.params = other.params
        self.view_options = other.view_options
        self.is_checked = other.is_checked
        self.is_preset = other.is_preset
        self.updated_at = other.updated_at
        return True


ChangeStatus = Literal[
    "added",
    "removed",
    "modified",
]
ChangeCategory = Literal[
    "breaking",
    "non_breaking",
    "partial_breaking",
    "unknown",
]


class NodeChange(BaseModel):
    category: ChangeCategory
    columns: Optional[dict[str, ChangeStatus]] = None


class NodeDiff(BaseModel):
    change_status: ChangeStatus
    change: Optional[NodeChange] = None  # Only available if change_status is 'modified'


class LineageDiff(BaseModel):
    base: dict
    current: dict
    diff: dict[str, NodeDiff]


# Column Level Linage
class CllColumnDep(BaseModel):
    node: str
    column: str


class CllColumn(BaseModel):
    id: Optional[str] = None
    table_id: Optional[str] = None
    name: Optional[str] = None

    # data type
    type: Optional[str] = None

    # transformation type
    transformation_type: Literal["source", "passthrough", "renamed", "derived", "unknown"] = "unknown"

    # change analysis
    change_status: Optional[ChangeStatus] = None

    # column-to-column dependencies
    depends_on: List[CllColumnDep] = Field(default_factory=list)


class CllNode(BaseModel):
    id: str
    name: str
    package_name: str
    resource_type: str
    raw_code: Optional[str] = None
    source_name: Optional[str] = None

    # change analysis
    change_status: Optional[ChangeStatus] = None
    change_category: Optional[ChangeCategory] = None

    # Column to column dependencies
    columns: Dict[str, CllColumn] = Field(default_factory=dict)

    # If the node is impacted. Only used if option 'change_analysis' is set
    impacted: Optional[bool] = None

    @classmethod
    def build_cll_node(cls, manifest, resource_key, node_id) -> Optional["CllNode"]:
        resources = getattr(manifest, resource_key)
        if node_id not in resources:
            return None
        n = resources[node_id]
        if resource_key == "nodes" and n.resource_type not in ["model", "seed", "snapshot"]:
            return None
        cll_node = CllNode(
            id=n.unique_id,
            name=n.name,
            package_name=n.package_name,
            resource_type=n.resource_type,
        )
        if resource_key == "sources":
            cll_node.source_name = n.source_name
        elif resource_key == "nodes":
            cll_node.raw_code = n.raw_code
        return cll_node


class CllData(BaseModel):
    nodes: Dict[str, CllNode] = Field(default_factory=dict)
    columns: Dict[str, CllColumn] = Field(default_factory=dict)
    parent_map: Dict[str, Set[str]] = Field(default_factory=dict)
    child_map: Dict[str, Set[str]] = Field(default_factory=dict)



import os
from unittest import TestCase
from unittest.mock import MagicMock

from recce.adapter.dbt_adapter import DbtAdapter, load_catalog, load_manifest

current_dir = os.path.dirname(os.path.abspath(__file__))


class TestAdapterLineage(TestCase):
    def setUp(self) -> None:
        self.manifest = load_manifest(path=os.path.join(current_dir, "manifest.json"))
        assert self.manifest is not None

        self.catalog = load_catalog(path=os.path.join(current_dir, "catalog.json"))
        assert self.catalog is not None

    def tearDown(self):
        pass

    def test_load_lineage(self):
        dbt_adapter = DbtAdapter(curr_manifest=self.manifest)
        lineage = dbt_adapter.get_lineage()
        assert lineage is not None
        assert lineage["nodes"]["model.jaffle_shop.orders"] is not None
        assert "columns" not in lineage["nodes"]["model.jaffle_shop.orders"]

    def test_load_lineage_with_catalog(self):
        mock_adapter = MagicMock()
        mock_adapter.type.return_value = None

        dbt_adapter = DbtAdapter(curr_manifest=self.manifest, curr_catalog=self.catalog)
        dbt_adapter.adapter = mock_adapter
        lineage = dbt_adapter.get_lineage()
        assert lineage is not None
        assert len(lineage["nodes"]["model.jaffle_shop.orders"]["columns"]) == 9



import time
from dataclasses import dataclass
from typing import Optional

import sqlglot.expressions as exp
from sqlglot import Dialect, parse_one
from sqlglot.errors import SqlglotError
from sqlglot.optimizer import Scope, traverse_scope
from sqlglot.optimizer.qualify import qualify

from recce.models.types import ChangeStatus, NodeChange

CHANGE_CATEGORY_UNKNOWN = NodeChange(category="unknown")
CHANGE_CATEGORY_BREAKING = NodeChange(category="breaking")


@dataclass
class BreakingPerformanceTracking:
    lineage_diff_start = None
    lineage_diff_elapsed = None
    modified_nodes = 0
    sqlglot_error_nodes = 0
    other_error_nodes = 0
    checkpoints = {}

    def start_lineage_diff(self):
        self.lineage_diff_start = time.perf_counter_ns()

    def record_checkpoint(self, label: str):
        if self.lineage_diff_start is None:
            return

        self.checkpoints[label] = (time.perf_counter_ns() - self.lineage_diff_start) / 1000000

    def end_lineage_diff(self):
        if self.lineage_diff_start is None:
            return
        self.lineage_diff_elapsed = (time.perf_counter_ns() - self.lineage_diff_start) / 1000000

    def increment_modified_nodes(self):
        self.modified_nodes += 1

    def increment_sqlglot_error_nodes(self):
        self.sqlglot_error_nodes += 1

    def increment_other_error_nodes(self):
        self.other_error_nodes += 1

    def to_dict(self):
        return {
            "lineage_diff_elapsed_ms": self.lineage_diff_elapsed,
            "modified_nodes": self.modified_nodes,
            "sqlglot_error_nodes": self.sqlglot_error_nodes,
            "other_error_nodes": self.other_error_nodes,
            "checkpoints": self.checkpoints,
        }

    def reset(self):
        self.lineage_diff_start = None
        self.lineage_diff_elapsed = None
        self.modified_nodes = 0
        self.sqlglot_error_nodes = 0
        self.other_error_nodes = 0
        self.checkpoints = {}


def _diff_select_scope(old_scope: Scope, new_scope: Scope, scope_changes_map: dict[Scope, NodeChange]) -> NodeChange:
    assert old_scope.expression.key == "select"
    assert new_scope.expression.key == "select"

    change_category = "non_breaking"
    changed_columns = {}

    # check if the upstream scopes is not breaking
    for source_name, source in new_scope.sources.items():
        if scope_changes_map.get(source) is not None:
            change = scope_changes_map[source]
            if change.category == "breaking":
                change_category = "breaking"

    # check if the upstream scopes sources table are the same
    if len(old_scope.sources) != len(new_scope.sources):
        change_category = "breaking"
    else:
        old_source_tables = [s.name for s in old_scope.sources.values() if isinstance(s, exp.Table)]
        new_source_tables = [s.name for s in new_scope.sources.values() if isinstance(s, exp.Table)]
        if sorted(old_source_tables) != sorted(new_source_tables):
            change_category = "breaking"

    # check if non-select expressions are the same
    old_select = old_scope.expression  # type: exp.Select
    new_select = new_scope.expression  # type: exp.Select
    for arg_key in old_select.args.keys() | new_select.args.keys():
        if arg_key in ["expressions", "with", "from"]:
            continue

        if old_select.args.get(arg_key) != new_select.args.get(arg_key):
            change_category = "breaking"

    def source_column_change_status(ref_column: exp.Column) -> Optional[ChangeStatus]:
        table_name = ref_column.table
        column_name = ref_column.name
        source = new_scope.sources.get(table_name, None)  # type: exp.Table | Scope
        if not isinstance(source, Scope):
            return None

        ref_change_category = scope_changes_map.get(source)
        if ref_change_category is None:
            return None

        return ref_change_category.columns.get(column_name)

    # selects
    old_column_map = {projection.alias_or_name: projection for projection in old_select.selects}
    new_column_map = {projection.alias_or_name: projection for projection in new_select.selects}
    is_distinct = new_select.args.get("distinct") is not None

    for column_name in old_column_map.keys() | new_column_map.keys():

        def _has_udtf(expr: exp.Expression) -> bool:
            return expr.find(exp.UDTF) is not None

        def _has_aggregate(expr: exp.Expression) -> bool:
            return expr.find(exp.AggFunc) is not None

        def _has_star(expr: exp.Expression) -> bool:
            return expr.find(exp.Star) is not None

        old_column = old_column_map.get(column_name)
        new_column = new_column_map.get(column_name)
        if old_column is None:
            if is_distinct:
                change_category = "breaking"
            elif _has_udtf(new_column):
                change_category = "breaking"

            changed_columns[column_name] = "added"
        elif new_column is None:
            if is_distinct:
                change_category = "breaking"
            elif _has_udtf(old_column):
                change_category = "breaking"

            changed_columns[column_name] = "removed"
            if change_category != "breaking":
                change_category = "partial_breaking"
        elif old_column != new_column:
            if is_distinct:
                change_category = "breaking"
            elif _has_udtf(old_column) and _has_udtf(new_column):
                change_category = "breaking"
            elif _has_aggregate(old_column) != _has_aggregate(new_column):
                change_category = "breaking"

            changed_columns[column_name] = "modified"
            if change_category != "breaking":
                change_category = "partial_breaking"
        else:
            if _has_star(new_column):
                for source_name, (_, source) in new_scope.selected_sources.items():
                    change = scope_changes_map.get(source)
                    if change is not None:
                        if change.category == "breaking":
                            change_category = "breaking"
                        for sub_column_name in change.columns.keys():
                            column_change_status = change.columns[sub_column_name]
                            changed_columns[sub_column_name] = column_change_status
                            if change_category != "breaking" and column_change_status in ["removed", "modified"]:
                                change_category = "partial_breaking"
                continue

            ref_columns = new_column.find_all(exp.Column)
            for ref_column in ref_columns:
                if source_column_change_status(ref_column) is not None:
                    if is_distinct:
                        change_category = "breaking"
                    elif _has_udtf(new_column):
                        change_category = "breaking"

                    if change_category != "breaking":
                        change_category = "partial_breaking"
                    changed_columns[column_name] = "modified"

    def selected_column_change_status(ref_column: exp.Column) -> Optional[ChangeStatus]:
        column_name = ref_column.name
        return changed_columns.get(column_name)

    # joins clause: Reference the source columns
    if new_select.args.get("joins"):
        joins = new_select.args.get("joins")
        for join in joins:
            if isinstance(join, exp.Join):
                for ref_column in join.find_all(exp.Column):
                    if source_column_change_status(ref_column) is not None:
                        change_category = "breaking"

    # where clauses: Reference the source columns
    if new_select.args.get("where"):
        where = new_select.args.get("where")
        if isinstance(where, exp.Where):
            for ref_column in where.find_all(exp.Column):
                if source_column_change_status(ref_column) is not None:
                    change_category = "breaking"

    # group by clause: Reference the source columns, column index
    if new_select.args.get("group"):
        group = new_select.args.get("group")
        if isinstance(group, exp.Group):
            for ref_column in group.find_all(exp.Column):
                if source_column_change_status(ref_column) is not None:
                    change_category = "breaking"

    # having clause: Reference the source columns, selected columns
    if new_select.args.get("having"):
        having = new_select.args.get("having")
        if isinstance(having, exp.Having):
            for ref_column in having.find_all(exp.Column):
                if source_column_change_status(ref_column) is not None:
                    change_category = "breaking"
                elif selected_column_change_status(ref_column) is not None:
                    change_category = "breaking"

    # order by clause: Reference the source columns, selected columns, column index
    if new_select.args.get("order"):
        order = new_select.args.get("order")
        if isinstance(order, exp.Order):
            for ref_column in order.find_all(exp.Column):
                if source_column_change_status(ref_column) is not None:
                    change_category = "breaking"
                elif selected_column_change_status(ref_column) is not None:
                    change_category = "breaking"

    return NodeChange(category=change_category, columns=changed_columns)


def _diff_union_scope(old_scope: Scope, new_scope: Scope, scope_changes_map: dict[Scope, NodeChange]) -> NodeChange:
    assert old_scope.expression.key == "union"
    assert new_scope.expression.key == "union"
    assert len(old_scope.union_scopes) == len(new_scope.union_scopes)
    assert new_scope.union_scopes is not None
    assert len(new_scope.union_scopes) > 0

    result_left = scope_changes_map.get(new_scope.union_scopes[0])
    change_category = result_left.category
    changed_columns = result_left.columns.copy()

    for sub_scope in new_scope.union_scopes[1:]:
        result_right = scope_changes_map.get(sub_scope)
        if change_category == "partial_breaking":
            if result_right.category in ["breaking"]:
                change_category = result_right.category
        elif change_category == "non_breaking":
            if result_right.category in ["breaking", "partial_breaking"]:
                change_category = result_right.category
        for column_name, column_change_status in result_right.columns.items():
            changed_columns[column_name] = column_change_status

    return NodeChange(category=change_category, columns=changed_columns)


def parse_change_category(
    old_sql,
    new_sql,
    old_schema=None,
    new_schema=None,
    dialect=None,
    perf_tracking: BreakingPerformanceTracking = None,
) -> NodeChange:
    if old_sql == new_sql:
        return NodeChange(category="non_breaking")

    try:
        dialect = Dialect.get(dialect)

        def _parse(sql, schema):
            exp = parse_one(sql, dialect=dialect)
            if schema:
                try:
                    exp = qualify(exp, schema=schema, dialect=dialect)
                except Exception:
                    # cannot optimize, skip it.
                    pass
            return exp

        old_exp = _parse(old_sql, old_schema)
        new_exp = _parse(new_sql, new_schema)
    except SqlglotError:
        if perf_tracking:
            perf_tracking.increment_sqlglot_error_nodes()
        return CHANGE_CATEGORY_UNKNOWN
    except Exception:
        if perf_tracking:
            perf_tracking.increment_other_error_nodes()
        return CHANGE_CATEGORY_UNKNOWN

    old_scopes = traverse_scope(old_exp)
    new_scopes = traverse_scope(new_exp)
    if len(old_scopes) != len(new_scopes):
        return NodeChange(category="breaking", columns={})

    scope_changes_map = {}
    for old_scope, new_scope in zip(old_scopes, new_scopes):
        if old_scope.expression.key != new_scope.expression.key:
            scope_changes_map[new_scope] = NodeChange(category="breaking")
            continue
        if old_scope == new_scope:
            scope_changes_map[new_scope] = NodeChange(category="non_breaking")
            continue

        scope_type = old_scope.expression.key
        if scope_type == "select":
            # CTE, Subquery, Root
            result = _diff_select_scope(old_scope, new_scope, scope_changes_map)
        elif scope_type == "union":
            # Union
            result = _diff_union_scope(old_scope, new_scope, scope_changes_map)
        else:
            if old_scope.expression != new_scope.expression:
                result = NodeChange(category="breaking", columns={})
            else:
                result = NodeChange(category="non_breaking", columns={})

        if result.category == "unknown":
            return result

        scope_changes_map[new_scope] = result
        if new_scope.is_root:
            return result

    return result



from .cloud import (
    CloudStateLoader,
    RecceCloudStateManager,
    RecceShareStateManager,
    s3_sse_c_headers,
)
from .const import ErrorMessage
from .local import FileStateLoader
from .state import (
    ArtifactsRoot,
    GitRepoInfo,
    PullRequestInfo,
    RecceState,
    RecceStateMetadata,
)
from .state_loader import RecceStateLoader

__all__ = [
    "ArtifactsRoot",
    "ErrorMessage",
    "RecceCloudStateManager",
    "RecceShareStateManager",
    "RecceState",
    "RecceStateLoader",
    "CloudStateLoader",
    "FileStateLoader",
    "RecceStateMetadata",
    "s3_sse_c_headers",
    "GitRepoInfo",
    "PullRequestInfo",
]



from recce.adapter.dbt_adapter import DbtAdapter
from recce.models.types import CllData
from recce.util.lineage import build_column_key


def assert_parent_map(result: CllData, node_or_column_id, parents):
    a_parents = result.parent_map.get(node_or_column_id) or set()

    assert len(a_parents) == len(parents), "parents length mismatch"
    for parent in parents:
        if isinstance(parent, str):
            node_id = parent
            assert node_id in a_parents, f"Node {node_id} not found in parent list"
        elif len(parent) == 1:
            (node_id,) = parent
            assert node_id in a_parents, f"Column {parent} not found in parent list"
        elif len(parent) == 2:
            node, column = parent
            column_id = build_column_key(node, column)
            assert column_id in a_parents, f"Column {column_id} not found in parent list for {node_or_column_id}"
        else:
            raise ValueError(f"Invalid parent format: {parent}. Expected node_id or (node_id, column_name).")


def assert_column(
    result: CllData,
    node_id,
    column_name,
    transformation_type=None,
    change_status=None,
    parents=None,
):
    column_id = build_column_key(node_id, column_name)
    entry = result.columns.get(column_id)
    assert entry is not None, f"Column {column_id} not found in result"
    assert (
        entry.transformation_type == transformation_type
    ), f"Column {column_name} type mismatch: expected {transformation_type}, got {entry.transformation_type}"
    assert_parent_map(result, column_id, parents)
    assert (
        entry.change_status == change_status
    ), f"Column {column_name} change status mismatch: expected {change_status}, got {entry.change_status}"


def assert_model(
    result: CllData,
    node_id,
    change_category=None,
    impacted=None,
    parents=None,
):
    entry = result.nodes.get(node_id)
    assert entry is not None, f"Node {node_id} not found in result"
    assert_parent_map(result, node_id, parents)

    assert (
        entry.change_category == change_category
    ), f"Node {node_id} change category mismatch: expected {change_category}, got {entry.change_category}"

    assert (entry.impacted == impacted), f"Node {node_id} impacted status mismatch: expected {impacted}, got {entry.impacted}"


def assert_cll_contain_nodes(cll_data: CllData, nodes):
    assert len(nodes) == len(cll_data.nodes), "Model count mismatch"
    for node in nodes:
        assert node in cll_data.nodes, f"Model {node} not found in lineage"


def assert_cll_contain_columns(cll_data: CllData, columns):
    assert len(columns) == len(cll_data.columns), "Column count mismatch"
    for column in columns:
        column_key = f"{column[0]}_{column[1]}"
        assert column_key in cll_data.columns, f"Column {column} not found in lineage"
        assert column[0] == cll_data.columns[column_key].table_id, f"Column {column[0]} node mismatch"
        assert column[1] == cll_data.columns[column_key].name, f"Column {column[1]} name mismatch"


def test_cll_basic(dbt_test_helper):
    dbt_test_helper.create_model(
        "model1", unique_id="model.model1", curr_sql="select 1 as c", curr_columns={"c": "int"}
    )
    dbt_test_helper.create_model(
        "model2",
        unique_id="model.model2",
        curr_sql='select c from {{ ref("model1") }} where c > 0',
        curr_columns={"c": "int"},
        depends_on=["model.model1"],
    )
    dbt_test_helper.create_model(
        "model3",
        unique_id="model.model3",
        curr_sql='select c from {{ ref("recce_test", "model1") }} where c > 0',
        curr_columns={"c": "int"},
        depends_on=["model.model1"],
    )
    adapter: DbtAdapter = dbt_test_helper.context.adapter

    result = adapter.get_cll("model.model1", "c")
    assert_model(result, "model.model2", parents=[("model.model1", "c")])
    assert_column(result, "model.model2", "c", transformation_type="passthrough", parents=[("model.model1", "c")])

    assert_model(result, "model.model3", parents=[("model.model1", "c")])
    assert_column(result, "model.model3", "c", transformation_type="passthrough", parents=[("model.model1", "c")])


def test_cll_table_alisa(dbt_test_helper):
    def patch_node(node):
        node["alias"] = "model1_alias"

    dbt_test_helper.create_model(
        "model1", unique_id="model.model1", curr_sql="select 1 as c", curr_columns={"c": "int"}, patch_func=patch_node
    )
    dbt_test_helper.create_model(
        "model2",
        unique_id="model.model2",
        curr_sql='select c from {{ ref("model1") }}',
        curr_columns={"c": "int"},
        depends_on=["model.model1"],
    )
    adapter: DbtAdapter = dbt_test_helper.context.adapter
    result = adapter.get_cll("model.model1", "c")
    assert_column(result, "model.model2", "c", transformation_type="passthrough", parents=[("model.model1", "c")])


def test_seed(dbt_test_helper):
    csv_data_curr = """
    customer_id,name,age
    1,Alice,30
    2,Bob,25
    3,Charlie,35
    """

    dbt_test_helper.create_model(
        "seed1",
        unique_id="seed.seed1",
        curr_csv=csv_data_curr,
        curr_columns={"customer_id": "varchar", "name": "varchar", "age": "int"},
        resource_type="seed",
    )
    dbt_test_helper.create_model(
        "model1",
        unique_id="model.model1",
        curr_sql='select customer_id from {{ ref("seed1") }} where age > 0',
        curr_columns={"customer_id": "varchar"},
        depends_on=["seed.seed1"],
    )
    adapter: DbtAdapter = dbt_test_helper.context.adapter

    result = adapter.get_cll("model.model1")
    assert_model(result, "seed.seed1", parents=[])
    assert_column(result, "seed.seed1", "customer_id", transformation_type="source", parents=[])
    assert_model(result, "model.model1", parents=["seed.seed1", ("seed.seed1", "age")])
    assert_column(
        result,
        "model.model1",
        "customer_id",
        transformation_type="passthrough",
        parents=[("seed.seed1", "customer_id")],
    )


def test_python_model(dbt_test_helper):
    def python_node(node):
        node["language"] = "python"

    csv_data_curr = """
    customer_id,name,age
    1,Alice,30
    2,Bob,25
    3,Charlie,35
    """
    dbt_test_helper.create_model(
        "model1", curr_csv=csv_data_curr, curr_columns={"customer_id": "varchar", "name": "varchar", "age": "int"}
    )
    dbt_test_helper.create_model(
        "model2",
        curr_csv=csv_data_curr,
        curr_columns={"customer_id": "varchar", "name": "varchar", "age": "int"},
        depends_on=["model1"],
        patch_func=python_node,
    )
    adapter: DbtAdapter = dbt_test_helper.context.adapter
    assert not adapter.is_python_model("model1")
    assert adapter.is_python_model("model2")

    result = adapter.get_cll("model2")
    assert_model(result, "model2", parents=["model1"])
    assert_column(result, "model2", "customer_id", transformation_type="unknown", parents=[])


def test_source(dbt_test_helper):
    csv_data_curr = """
        customer_id,name,age
        1,Alice,30
        2,Bob,25
        3,Charlie,35
        """

    dbt_test_helper.create_source(
        "source1",
        "table1",
        unique_id="source.source1.table1",
        curr_csv=csv_data_curr,
        curr_columns={"customer_id": "varchar", "name": "varchar", "age": "int"},
    )
    dbt_test_helper.create_model(
        "model1",
        unique_id="model.model1",
        curr_sql='select customer_id from {{ source("source1", "table1") }}',
        curr_columns={"customer_id": "int"},
        depends_on=["source.source1.table1"],
    )
    adapter: DbtAdapter = dbt_test_helper.context.adapter
    result = adapter.get_cll("model.model1")
    assert_column(result, "source.source1.table1", "customer_id", transformation_type="source", parents=[])
    assert_column(
        result,
        "model.model1",
        "customer_id",
        transformation_type="passthrough",
        parents=[("source.source1.table1", "customer_id")],
    )


def test_parse_error(dbt_test_helper):
    dbt_test_helper.create_model("model1", curr_sql="select 1 as c", curr_columns={"c": "int"})
    dbt_test_helper.create_model("model2", curr_sql="this is not a valid sql", curr_columns={"c": "int"})
    adapter: DbtAdapter = dbt_test_helper.context.adapter
    result = adapter.get_cll("model2")
    assert_column(result, "model2", "c", transformation_type="unknown", parents=[])


def test_model_without_catalog(dbt_test_helper):
    dbt_test_helper.create_model("model1", curr_sql="select 1 as c")
    adapter: DbtAdapter = dbt_test_helper.context.adapter
    result = adapter.get_cll("model1")
    assert not result.nodes["model1"].columns


def test_column_level_lineage(dbt_test_helper):
    dbt_test_helper.create_model(
        "model1", unique_id="model.model1", curr_sql="select 1 as c", curr_columns={"c": "int"}
    )
    dbt_test_helper.create_model(
        "model2",
        unique_id="model.model2",
        curr_sql='select c, 2025 as y from {{ ref("model1") }}',
        curr_columns={"c": "int", "y": "int"},
        depends_on=["model.model1"],
    )
    dbt_test_helper.create_model(
        "model3",
        unique_id="model.model3",
        curr_sql='select c from {{ ref("model2") }} where y < 2025',
        curr_columns={"c": "int"},
        depends_on=["model.model2"],
    )
    dbt_test_helper.create_model(
        "model4",
        unique_id="model.model4",
        curr_sql='select y from {{ ref("model2") }}',
        curr_columns={"y": "int"},
        depends_on=["model.model2"],
    )

    adapter: DbtAdapter = dbt_test_helper.context.adapter

    result = adapter.get_cll("model.model2", "c")
    assert_cll_contain_nodes(result, [])
    assert_cll_contain_columns(result, [("model.model1", "c"), ("model.model2", "c"), ("model.model3", "c")])
    assert_column(result, "model.model2", "c", transformation_type="passthrough", parents=[("model.model1", "c")])

    result = adapter.get_cll("model.model2", "y")
    assert_cll_contain_nodes(result, ["model.model3"])
    assert_cll_contain_columns(result, [("model.model2", "y"), ("model.model4", "y")])
    assert_column(result, "model.model2", "y", transformation_type="source", parents=[])

    result = adapter.get_cll("model.model3", "c")
    assert_cll_contain_nodes(result, [])
    assert_cll_contain_columns(result, [("model.model1", "c"), ("model.model2", "c"), ("model.model3", "c")])
    assert_column(result, "model.model2", "c", transformation_type="passthrough", parents=[("model.model1", "c")])

    result = adapter.get_cll("model.model2", "c", no_upstream=True, no_downstream=True)
    assert_cll_contain_nodes(result, [])
    assert_cll_contain_columns(result, [("model.model2", "c")])
    assert_column(result, "model.model2", "c", transformation_type="passthrough", parents=[])


def test_impact_radius_no_change_analysis_no_cll(dbt_test_helper):
    dbt_test_helper.create_model(
        "model1",
        unique_id="model.model1",
        curr_sql="select 1 as c",
        base_sql="select 1 as c --- non-breaking",
        curr_columns={"c": "int"},
        base_columns={"c": "int"},
    )
    dbt_test_helper.create_model(
        "model2",
        unique_id="model.model2",
        curr_sql='select c, 2025 as y from {{ ref("model1") }}',
        base_sql='select c, 2025 as y from {{ ref("model1") }} where c > 0 --- breaking',
        curr_columns={"c": "int", "y": "int"},
        base_columns={"c": "int", "y": "int"},
        depends_on=["model.model1"],
    )
    dbt_test_helper.create_model(
        "model3",
        unique_id="model.model3",
        curr_sql='select c from {{ ref("model2") }} where y < 2025',
        base_sql='select c from {{ ref("model2") }} where y < 2025',
        curr_columns={"c": "int"},
        base_columns={"c": "int"},
        depends_on=["model.model2"],
    )
    dbt_test_helper.create_model(
        "model4",
        unique_id="model.model4",
        curr_sql='select y + 1 as year from {{ ref("model2") }} --- partial breaking',
        base_sql='select y as year from {{ ref("model2") }}',
        curr_columns={"year": "int"},
        base_columns={"year": "int"},
        depends_on=["model.model2"],
    )
    dbt_test_helper.create_model(
        "model5",
        unique_id="model.model5",
        curr_sql='select c, 2025 as y from {{ ref("model1") }}',
        base_sql='select c, 2025 as y from {{ ref("model1") }}',
        curr_columns={"c": "int", "y": "int"},
        base_columns={"c": "int", "y": "int"},
        depends_on=["model.model1"],
    )

    adapter: DbtAdapter = dbt_test_helper.context.adapter

    result = adapter.get_cll(no_cll=True)
    assert_cll_contain_nodes(result, ["model.model1", "model.model2", "model.model3", "model.model4", "model.model5"])
    assert_model(result, "model.model1", parents=[])
    assert_model(result, "model.model2", parents=["model.model1"])
    assert_model(result, "model.model3", parents=["model.model2"])
    assert_model(result, "model.model4", parents=["model.model2"])
    assert_model(result, "model.model5", parents=["model.model1"])


def test_impact_radius_with_change_analysis_no_cll(dbt_test_helper):
    dbt_test_helper.create_model(
        "model1",
        unique_id="model.model1",
        curr_sql="select 1 as c",
        base_sql="select 1 as c --- non-breaking",
        curr_columns={"c": "int"},
        base_columns={"c": "int"},
    )
    dbt_test_helper.create_model(
        "model2",
        unique_id="model.model2",
        curr_sql='select c, 2025 as y from {{ ref("model1") }}',
        base_sql='select c, 2025 as y from {{ ref("model1") }} where c > 0 --- breaking',
        curr_columns={"c": "int", "y": "int"},
        base_columns={"c": "int", "y": "int"},
        depends_on=["model.model1"],
    )
    dbt_test_helper.create_model(
        "model3",
        unique_id="model.model3",
        curr_sql='select c from {{ ref("model2") }} where y < 2025',
        base_sql='select c from {{ ref("model2") }} where y < 2025',
        curr_columns={"c": "int"},
        base_columns={"c": "int"},
        depends_on=["model.model2"],
    )
    dbt_test_helper.create_model(
        "model4",
        unique_id="model.model4",
        curr_sql='select y + 1 as year from {{ ref("model2") }} --- partial breaking',
        base_sql='select y as year from {{ ref("model2") }}',
        curr_columns={"year": "int"},
        base_columns={"year": "int"},
        depends_on=["model.model2"],
    )
    dbt_test_helper.create_model(
        "model5",
        unique_id="model.model5",
        curr_sql='select c, 2025 as y from {{ ref("model1") }}',
        base_sql='select c, 2025 as y from {{ ref("model1") }}',
        curr_columns={"c": "int", "y": "int"},
        base_columns={"c": "int", "y": "int"},
        depends_on=["model.model1"],
    )

    adapter: DbtAdapter = dbt_test_helper.context.adapter

    # breaking
    result = adapter.get_cll(change_analysis=True, no_cll=True, no_upstream=True)
    assert_cll_contain_nodes(result, ["model.model1", "model.model2", "model.model3", "model.model4"])
    assert_model(result, "model.model1", parents=[], change_category="non_breaking", impacted=False)
    assert_model(result, "model.model2", parents=[], change_category="breaking", impacted=True)
    assert_model(result, "model.model3", parents=["model.model2"], change_category=None, impacted=True)
    assert_model(result, "model.model4", parents=["model.model2"], change_category="partial_breaking", impacted=True)


def test_impact_radius_with_change_analysis_no_cll_2(dbt_test_helper):
    # partial breaking
    dbt_test_helper.create_model(
        "model1",
        unique_id="model.model1",
        curr_sql="select 1 as c",
        base_sql="select 2 as c",
        curr_columns={"c": "int"},
        base_columns={"c": "int"},
    )
    # breaking
    dbt_test_helper.create_model(
        "model2",
        unique_id="model.model2",
        curr_sql='select c, 2025 as y from {{ ref("model1") }}',
        base_sql='select c, 2025 as y from {{ ref("model1") }} where c > 0 --- breaking',
        curr_columns={"c": "int", "y": "int"},
        base_columns={"c": "int", "y": "int"},
        depends_on=["model.model1"],
    )
    # no change
    dbt_test_helper.create_model(
        "model3",
        unique_id="model.model3",
        curr_sql='select c from {{ ref("model2") }} where y < 2025',
        base_sql='select c from {{ ref("model2") }} where y < 2025',
        curr_columns={"c": "int"},
        base_columns={"c": "int"},
        depends_on=["model.model2"],
    )
    # partial breaking
    dbt_test_helper.create_model(
        "model4",
        unique_id="model.model4",
        curr_sql='select y + 1 as year from {{ ref("model2") }} --- partial breaking',
        base_sql='select y as year from {{ ref("model2") }}',
        curr_columns={"year": "int"},
        base_columns={"year": "int"},
        depends_on=["model.model2"],
    )
    # no change
    dbt_test_helper.create_model(
        "model5",
        unique_id="model.model5",
        curr_sql='select c, 2025 as y from {{ ref("model1") }}',
        base_sql='select c, 2025 as y from {{ ref("model1") }}',
        curr_columns={"c": "int", "y": "int"},
        base_columns={"c": "int", "y": "int"},
        depends_on=["model.model1"],
    )

    adapter: DbtAdapter = dbt_test_helper.context.adapter

    # breaking
    result = adapter.get_cll(change_analysis=True, no_cll=True, no_upstream=True)
    assert_cll_contain_nodes(result, ["model.model1", "model.model2", "model.model3", "model.model4", "model.model5"])
    assert_model(result, "model.model1", parents=[], change_category="partial_breaking", impacted=True)
    assert_model(result, "model.model2", parents=["model.model1"], change_category="breaking", impacted=True)
    assert_model(result, "model.model3", parents=["model.model2"], change_category=None, impacted=True)
    assert_model(result, "model.model4", parents=["model.model2"], change_category="partial_breaking", impacted=True)
    assert_model(result, "model.model5", parents=["model.model1"], impacted=True)


def test_impact_radius_with_change_analysis_with_cll(dbt_test_helper):
    dbt_test_helper.create_model(
        "model1",
        unique_id="model.model1",
        curr_sql="select 1 as c",
        base_sql="select 1 as c --- non-breaking",
        curr_columns={"c": "int"},
        base_columns={"c": "int"},
    )
    dbt_test_helper.create_model(
        "model2",
        unique_id="model.model2",
        curr_sql='select c, 2025 as y from {{ ref("model1") }}',
        base_sql='select c, 2025 as y from {{ ref("model1") }} where c > 0 --- breaking',
        curr_columns={"c": "int", "y": "int"},
        base_columns={"c": "int", "y": "int"},
        depends_on=["model.model1"],
    )
    dbt_test_helper.create_model(
        "model3",
        unique_id="model.model3",
        curr_sql='select c from {{ ref("model2") }} where y < 2025',
        base_sql='select c from {{ ref("model2") }} where y < 2025',
        curr_columns={"c": "int"},
        base_columns={"c": "int"},
        depends_on=["model.model2"],
    )
    dbt_test_helper.create_model(
        "model4",
        unique_id="model.model4",
        curr_sql='select y + 1 as year from {{ ref("model2") }} --- partial breaking',
        base_sql='select y as year from {{ ref("model2") }}',
        curr_columns={"year": "int"},
        base_columns={"year": "int"},
        depends_on=["model.model2"],
    )
    dbt_test_helper.create_model(
        "model5",
        unique_id="model.model5",
        curr_sql='select c, 2025 as y from {{ ref("model1") }}',
        base_sql='select c, 2025 as y from {{ ref("model1") }}',
        curr_columns={"c": "int", "y": "int"},
        base_columns={"c": "int", "y": "int"},
        depends_on=["model.model1"],
    )

    adapter: DbtAdapter = dbt_test_helper.context.adapter

    result = adapter.get_cll(change_analysis=True, no_upstream=True)
    assert_cll_contain_nodes(result, ["model.model1", "model.model2", "model.model3", "model.model4"])
    assert_cll_contain_columns(result, [("model.model4", "year")])
    assert_model(result, "model.model1", parents=[], change_category="non_breaking", impacted=False)
    assert_model(result, "model.model2", parents=[], change_category="breaking", impacted=True)
    assert_model(result, "model.model3", parents=["model.model2"], change_category=None, impacted=True)
    assert_model(result, "model.model4", parents=["model.model2"], change_category="partial_breaking", impacted=True)


def test_impact_radius_with_change_analysis_with_cll_added_removed(dbt_test_helper):
    # rename model
    dbt_test_helper.create_model(
        "model1",
        unique_id="model.model1",
        base_sql="select 1 as c",
        base_columns={"c": "int"},
    )
    dbt_test_helper.create_model(
        "model1_v2",
        unique_id="model.model1_v2",
        curr_sql="select 1 as c",
        curr_columns={"c": "int"},
    )
    # change upstream
    dbt_test_helper.create_model(
        "model2",
        unique_id="model.model2",
        base_sql='select c from {{ ref("model1") }}',
        base_columns={"c": "int"},
        depends_on=["model.model1"],
    )
    dbt_test_helper.create_model(
        "model2",
        unique_id="model.model2",
        curr_sql='select c from {{ ref("model1_v2") }}',
        curr_columns={"c": "int"},
        depends_on=["model.model1_v2"],
    )

    adapter: DbtAdapter = dbt_test_helper.context.adapter
    result = adapter.get_cll(change_analysis=True, no_upstream=True)
    assert_model(result, "model.model1_v2", parents=[], impacted=True)
    assert_column(result, "model.model1_v2", "c", transformation_type="source", change_status="added", parents=[])
    assert_model(result, "model.model2", parents=["model.model1_v2"], change_category="breaking", impacted=True)
    assert_cll_contain_nodes(result, ["model.model1_v2", "model.model2"])
    assert_cll_contain_columns(result, [("model.model1_v2", "c"), ("model.model2", "c")])


def test_impact_radius_by_node_no_cll(dbt_test_helper):
    # non-breaking
    dbt_test_helper.create_model(
        "model1",
        unique_id="model.model1",
        curr_sql="select 1 as c",
        base_sql="select 1 as c --- non-breaking",
        curr_columns={"c": "int"},
        base_columns={"c": "int"},
    )
    # breaking
    dbt_test_helper.create_model(
        "model2",
        unique_id="model.model2",
        curr_sql='select c, 2025 as y from {{ ref("model1") }}',
        base_sql='select c, 2025 as y from {{ ref("model1") }} where c > 0 --- breaking',
        curr_columns={"c": "int", "y": "int"},
        base_columns={"c": "int", "y": "int"},
        depends_on=["model.model1"],
    )
    dbt_test_helper.create_model(
        "model3",
        unique_id="model.model3",
        curr_sql='select c from {{ ref("model2") }} where y < 2025',
        base_sql='select c from {{ ref("model2") }} where y < 2025',
        curr_columns={"c": "int"},
        base_columns={"c": "int"},
        depends_on=["model.model2"],
    )
    dbt_test_helper.create_model(
        "model4",
        unique_id="model.model4",
        curr_sql='select y from {{ ref("model2") }}',
        base_sql='select y from {{ ref("model2") }}',
        curr_columns={"y": "int"},
        base_columns={"y": "int"},
        depends_on=["model.model2"],
    )

    adapter: DbtAdapter = dbt_test_helper.context.adapter

    # breaking
    result = adapter.get_cll(node_id="model.model2", change_analysis=True, no_cll=True, no_upstream=True)
    assert_cll_contain_nodes(result, ["model.model2", "model.model3", "model.model4"])

    # non-breaking
    result = adapter.get_cll(node_id="model.model1", change_analysis=True, no_cll=True, no_upstream=True)
    assert_cll_contain_nodes(result, ["model.model1"])


def test_impact_radius_by_node_with_cll(dbt_test_helper):
    # added column
    dbt_test_helper.create_model(
        "model1",
        unique_id="model.model1",
        curr_sql="select 1 as c, 2 as d --- add d",
        base_sql="select 1 as c",
        curr_columns={"c": "int", "d": "int"},
        base_columns={"c": "int"},
    )
    # modified column
    dbt_test_helper.create_model(
        "model2",
        unique_id="model.model2",
        curr_sql='select c, 2024 as y from {{ ref("model1") }} --- modify y',
        base_sql='select c, 2025 as y from {{ ref("model1") }}',
        curr_columns={"c": "int", "y": "int"},
        base_columns={"c": "int", "y": "int"},
        depends_on=["model.model1"],
    )
    dbt_test_helper.create_model(
        "model3",
        unique_id="model.model3",
        curr_sql='select c from {{ ref("model2") }} where y < 2025',
        base_sql='select c from {{ ref("model2") }} where y < 2025',
        curr_columns={"c": "int"},
        base_columns={"c": "int"},
        depends_on=["model.model2"],
    )
    dbt_test_helper.create_model(
        "model4",
        unique_id="model.model4",
        curr_sql='select y from {{ ref("model2") }}',
        base_sql='select y from {{ ref("model2") }}',
        curr_columns={"y": "int"},
        base_columns={"y": "int"},
        depends_on=["model.model2"],
    )

    adapter: DbtAdapter = dbt_test_helper.context.adapter

    result = adapter.get_cll(node_id="model.model2", change_analysis=True, no_upstream=True)
    assert_model(result, "model.model2", parents=[], change_category="partial_breaking", impacted=False)
    assert_model(result, "model.model3", parents=[('model.model2', 'y')], impacted=True)
    assert_column(result, "model.model2", "y", transformation_type="source", parents=[], change_status="modified")
    assert_cll_contain_nodes(result, ["model.model2", "model.model3"])
    assert_cll_contain_columns(result, [("model.model2", "y"), ("model.model4", "y")])

    result = adapter.get_cll(node_id="model.model1", change_analysis=True, no_upstream=True)
    assert_cll_contain_nodes(result, ["model.model1"])
    assert_cll_contain_columns(result, [("model.model1", "d")])
    assert_model(result, "model.model1", parents=[], change_category="non_breaking", impacted=False)
    assert_column(result, "model.model1", "d", transformation_type="source", parents=[], change_status="added")


def test_impact_radius_by_node_with_cll_2(dbt_test_helper):
    # added column
    dbt_test_helper.create_model(
        "model1",
        unique_id="model.model1",
        curr_sql="select 1 as c, 2 as d --- add d",
        base_sql="select 1 as c",
        curr_columns={"c": "int", "d": "int"},
        base_columns={"c": "int"},
    )
    # breaking, added column, modified column
    dbt_test_helper.create_model(
        "model2",
        unique_id="model.model2",
        curr_sql='select c, 2024 as y, d from {{ ref("model1") }} --- modify y, add d',
        base_sql='select c, 2025 as y from {{ ref("model1") }} where c > 0 --- breaking',
        curr_columns={"c": "int", "y": "int", "d": "int"},
        base_columns={"c": "int", "y": "int"},
        depends_on=["model.model1"],
    )
    dbt_test_helper.create_model(
        "model3",
        unique_id="model.model3",
        curr_sql='select c from {{ ref("model2") }} where y < 2025',
        base_sql='select c from {{ ref("model2") }} where y < 2025',
        curr_columns={"c": "int"},
        base_columns={"c": "int"},
        depends_on=["model.model2"],
    )
    dbt_test_helper.create_model(
        "model4",
        unique_id="model.model4",
        curr_sql='select y from {{ ref("model2") }}',
        base_sql='select y from {{ ref("model2") }}',
        curr_columns={"y": "int"},
        base_columns={"y": "int"},
        depends_on=["model.model2"],
    )

    adapter: DbtAdapter = dbt_test_helper.context.adapter

    result = adapter.get_cll(node_id="model.model2", change_analysis=True, no_upstream=True)
    assert_model(result, "model.model2", parents=[], change_category="breaking", impacted=True)
    assert_column(result, "model.model2", "y", transformation_type="source", parents=[], change_status="modified")
    assert_model(result, "model.model3", parents=["model.model2", ("model.model2", "y")], impacted=True)
    assert_model(result, "model.model4", parents=["model.model2"], impacted=True)
    assert_column(result, "model.model4", "y", transformation_type="passthrough", parents=[("model.model2", "y")])
    assert_cll_contain_nodes(result, ["model.model2", "model.model3", "model.model4"])
    assert_cll_contain_columns(result, [("model.model2", "d"), ("model.model2", "y"), ("model.model4", "y")])



import time
from dataclasses import dataclass


@dataclass
class LineagePerfTracker:
    lineage_start = None
    lineage_elapsed = None
    column_lineage_start = None
    column_lineage_elapsed = None

    total_nodes = None
    init_nodes = None
    cll_nodes = 0
    change_analysis_nodes = 0
    anchor_nodes = None

    params = None

    def start_lineage(self):
        self.lineage_start = time.perf_counter_ns()

    def end_lineage(self):
        if self.lineage_start is None:
            return
        self.lineage_elapsed = (time.perf_counter_ns() - self.lineage_start) / 1000000

    def start_column_lineage(self):
        self.column_lineage_start = time.perf_counter_ns()

    def end_column_lineage(self):
        if self.column_lineage_start is None:
            return
        self.column_lineage_elapsed = (time.perf_counter_ns() - self.column_lineage_start) / 1000000

    def set_total_nodes(self, total_nodes):
        self.total_nodes = total_nodes

    def set_init_nodes(self, init_nodes):
        self.init_nodes = init_nodes

    def set_anchor_nodes(self, anchor_nodes):
        self.anchor_nodes = anchor_nodes

    def increment_cll_nodes(self):
        self.cll_nodes += 1

    def increment_change_analysis_nodes(self):
        self.change_analysis_nodes += 1

    def set_params(self, has_node, has_column, change_analysis, no_cll, no_upstream, no_downstream):
        self.params = {
            "has_node": has_node,
            "has_column": has_column,
            "change_analysis": change_analysis,
            "no_cll": no_cll,
            "no_upstream": no_upstream,
            "no_downstream": no_downstream,
        }

    def to_dict(self):
        return {
            "lineage_elapsed_ms": self.lineage_elapsed,
            "column_lineage_elapsed_ms": self.column_lineage_elapsed,
            "total_nodes": self.total_nodes,
            "init_nodes": self.init_nodes,
            "cll_nodes": self.cll_nodes,
            "change_analysis_nodes": self.change_analysis_nodes,
            "anchor_nodes": self.anchor_nodes,
            "params": self.params,
        }

    def reset(self):
        self.lineage_start = None
        self.lineage_elapsed = None
        self.column_lineage_start = None
        self.column_lineage_elapsed = None

        self.total_nodes = None
        self.init_nodes = None
        self.change_analysis_nodes = 0
        self.cll_nodes = 0
        self.anchor_nodes = 0

        self.params = None



class DbtVersion:

    def __init__(self):
        from dbt import version as dbt_version

        dbt_version = self.parse(dbt_version.__version__)
        if dbt_version.is_prerelease:
            dbt_version = self.parse(dbt_version.base_version)
        self.dbt_version = dbt_version

    @staticmethod
    def parse(version: str):
        from packaging import version as v

        return v.parse(version)

    def as_version(self, other):
        from packaging.version import Version

        if isinstance(other, Version):
            return other
        if isinstance(other, str):
            return self.parse(other)
        return self.parse(str(other))

    def __ge__(self, other):
        return self.dbt_version >= self.as_version(other)

    def __gt__(self, other):
        return self.dbt_version > self.as_version(other)

    def __lt__(self, other):
        return self.dbt_version < self.as_version(other)

    def __le__(self, other):
        return self.dbt_version <= self.as_version(other)

    def __eq__(self, other):
        return self.dbt_version.release[:2] == self.as_version(other).release[:2]

    def __str__(self):
        return ".".join([str(x) for x in list(self.dbt_version.release)])



import gzip
import os
import tempfile
from abc import ABC, ABCMeta, abstractmethod
from enum import Enum


class SupportedFileTypes(Enum):
    FILE = "file"
    GZIP = "gzip"
    ZIP = "zip"


def file_io_factory(file_type: SupportedFileTypes):
    if file_type == SupportedFileTypes.FILE:
        return FileIO
    elif file_type == SupportedFileTypes.GZIP:
        return GzipFileIO
    elif file_type == SupportedFileTypes.ZIP:
        return ZipFileIO
    else:
        raise ValueError(f"Unsupported file type: {file_type}")


class AbstractFileIO(metaclass=ABCMeta):
    @staticmethod
    @abstractmethod
    def write(path: str, data: str, **kwargs):
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def read(path: str, **kwargs) -> str:
        raise NotImplementedError


class FileIO(AbstractFileIO, ABC):
    @staticmethod
    def write(path: str, data: str, **kwargs):
        with open(path, "w", encoding="utf-8") as f:
            f.write(data)

    @staticmethod
    def read(path: str, **kwargs) -> str:
        with open(path, "r", encoding="utf-8") as f:
            return f.read()


class GzipFileIO(AbstractFileIO, ABC):
    @staticmethod
    def write(path: str, data: str, **kwargs):
        with gzip.open(path, "wt") as f:
            f.write(data)

    @staticmethod
    def read(path: str, **kwargs) -> str:
        with gzip.open(path, "rt") as f:
            return f.read()

    @staticmethod
    def read_fileobj(fileobj) -> bytes:
        with gzip.GzipFile(fileobj=fileobj) as f:
            return f.read()


class ZipFileIO(AbstractFileIO, ABC):
    @staticmethod
    def _is_pyminizip_installed():
        try:
            import pyminizip

            # Use the module to avoid F401
            return pyminizip is not None
        except ImportError:
            raise ImportError("pyminizip is not installed. Please install it using `pip install pyminizip`")

    @staticmethod
    def read(path: str, **kwargs) -> str:
        ZipFileIO._is_pyminizip_installed()
        import pyminizip

        cwd = os.getcwd()
        password = kwargs.get("password")
        zip_dir_name = kwargs.get("zip_dir_name")
        if zip_dir_name is None:
            raise ValueError("zip_dir_name is required for zipping")

        try:
            with tempfile.TemporaryDirectory() as tmp_dir:
                tmp_file = os.path.join(tmp_dir, zip_dir_name)
                pyminizip.uncompress(path, password, tmp_dir, 0)
                content = FileIO.read(tmp_file)
        except Exception as e:
            error_msg = str(e)
            if "-3" in error_msg:
                raise Exception("Invalid password to uncompress state file.")
            raise Exception(f"Failed to uncompress state file: {error_msg}")
        finally:
            os.chdir(cwd)

        return content

    @staticmethod
    def write(path: str, data: str, **kwargs):
        ZipFileIO._is_pyminizip_installed()
        import pyminizip

        cwd = os.getcwd()
        password = kwargs.get("password")
        zip_dir_name = kwargs.get("zip_dir_name")
        if zip_dir_name is None:
            raise ValueError("zip_dir_name is required for zipping")

        try:
            with tempfile.TemporaryDirectory() as tmp_dir:
                tmp_file = os.path.join(tmp_dir, zip_dir_name)
                FileIO.write(tmp_file, data)
                pyminizip.compress(tmp_file, None, path, password, 9)
        finally:
            os.chdir(cwd)



class RecceException(Exception):
    def __init__(self, message, is_raise=True):
        self.message = message
        self.is_raise = is_raise
        super().__init__(self.message)


class RecceCancelException(RecceException):
    def __init__(self):
        super().__init__("Cancelled", False)


class RecceConfigException(RecceException):
    def __init__(self, message, cause=None):
        super().__init__(message)
        self.cause = cause

    def __str__(self):
        if self.cause:
            return f"{super().__str__()}\n{str(self.cause)}"
        return super().__str__()



from .core import Task
from .histogram import HistogramDiffTask
from .profile import ProfileDiffTask, ProfileTask
from .query import QueryBaseTask, QueryDiffTask, QueryTask
from .rowcount import RowCountDiffTask, RowCountTask
from .top_k import TopKDiffTask
from .valuediff import ValueDiffDetailTask, ValueDiffTask

# Explicitly declare exports
__all__ = [
    "Task",
    "HistogramDiffTask",
    "ProfileDiffTask",
    "ProfileTask",
    "QueryBaseTask",
    "QueryDiffTask",
    "QueryTask",
    "RowCountDiffTask",
    "RowCountTask",
    "TopKDiffTask",
    "ValueDiffDetailTask",
    "ValueDiffTask",
]



# noinspection PyUnresolvedReferences
import os
import unittest
from datetime import datetime

from recce.core import RecceContext
from recce.models import Check, Run, RunType
from recce.state import ArtifactsRoot, FileStateLoader, RecceState
from tests.adapter.dbt_adapter.conftest import dbt_test_helper  # noqa: F401

current_dir = os.path.dirname(os.path.abspath(__file__))


class TestRecceState(unittest.TestCase):
    def test_load(self):
        run = Run(type=RunType.QUERY, params=dict(sql_template="select * from users"))
        check = Check(name="check 1", description="desc 1", type=run.type, params=run.params)

        state = RecceState(runs=[run], checks=[check])
        json_content = state.to_json()
        new_state = RecceState.from_json(json_content)

        run_loaded = new_state.runs[0]
        check_loaded = new_state.checks[0]

        assert run.run_id == run_loaded.run_id
        assert check.check_id == check_loaded.check_id

    def test_merge_checks(self):
        check1 = Check(name="test1", description="", type="query")
        check2 = Check(name="test2", description="", type="query", updated_at=datetime(2000, 1, 1))
        check2_2 = Check(
            name="test2_2", description="", type="query", updated_at=datetime(2020, 1, 1), check_id=check2.check_id
        )
        check3 = Check(name="test3", description="", type="query")

        context = RecceContext()
        state = RecceState(checks=[check1], runs=[])
        context.import_state(state)
        self.assertEqual(1, len(context.checks))
        self.assertEqual(check1.name, context.checks[0].name)

        context = RecceContext(checks=[check1, check2])
        state = RecceState(checks=[check1, check2_2, check3], runs=[])
        context.import_state(state)
        self.assertEqual(3, len(context.checks))
        self.assertEqual(check2_2.name, context.checks[1].name)

    def test_merge_preset_checks(self):
        check1 = Check(
            name="test1",
            description="test1",
            type="query",
            params=dict(foo="bar"),
            updated_at=datetime(2000, 1, 1),
            is_preset=True,
        )
        check2 = Check(
            name="test2",
            description="test2",
            type="query",
            params=dict(foo="bar"),
            updated_at=datetime(2001, 1, 1),
            is_preset=True,
        )

        context = RecceContext(checks=[check1])
        state = RecceState(checks=[check2], runs=[])
        context.import_state(state)
        self.assertEqual(1, len(context.checks))
        self.assertEqual(check2.name, context.checks[0].name)

        context = RecceContext(checks=[check2])
        state = RecceState(checks=[check1], runs=[])
        context.import_state(state)
        self.assertEqual(1, len(context.checks))
        self.assertEqual(check2.name, context.checks[0].name)

    def test_revert_checks(self):
        check1 = Check(name="test1", description="", type="query")
        check2 = Check(name="test2", description="", type="query")
        check2_2 = Check(name="test2_2", description="", type="query", check_id=check2.check_id)
        check3 = Check(name="test3", description="", type="query")

        context = RecceContext(checks=[check1, check2])
        state = RecceState(checks=[check2_2, check3], runs=[])
        context.import_state(state, merge=False)
        self.assertEqual(2, len(context.checks))
        self.assertEqual(check2_2.name, context.checks[0].name)

    def test_merge_runs(self):
        run1 = Run(type="query")
        run2 = Run(type="query")
        run3 = Run(type="query")

        context = RecceContext(runs=[])
        state = RecceState(runs=[run1])
        context.import_state(state)
        self.assertEqual(1, len(context.runs))

        context = RecceContext(runs=[run1, run2])
        state = RecceState(runs=[run2, run3])
        context.import_state(state)
        self.assertEqual(3, len(context.runs))

    def test_merge_dbt_artifacts(self):
        import json
        import os

        with open(os.path.join(current_dir, "manifest.json"), "r") as f:
            manifest = json.load(f)
        manifest["metadata"]["generated_at"] = "2000-01-01T00:00:00Z"
        artifacts = ArtifactsRoot(
            base=dict(
                manifest=manifest,
            ),
            current=dict(
                manifest=manifest,
            ),
        )

        from tests.adapter.dbt_adapter.dbt_test_helper import DbtTestHelper

        adapter = DbtTestHelper().adapter
        adapter.import_artifacts(artifacts)
        self.assertNotEqual(adapter.base_manifest.metadata.invocation_id, manifest.get("metadata").get("invocation_id"))

        manifest["metadata"]["generated_at"] = "2099-01-01T00:00:00Z"
        adapter.import_artifacts(artifacts)
        self.assertEqual(adapter.base_manifest.metadata.invocation_id, manifest.get("metadata").get("invocation_id"))

    def test_state_loader(self):
        # copy ./recce_state.json to temp and open

        # use library to create a temp file in the context
        import os
        import shutil
        import tempfile

        with tempfile.NamedTemporaryFile() as f:
            # copy ./recce_state.json to temp file
            current_dir = os.path.dirname(os.path.abspath(__file__))
            state_file = os.path.join(current_dir, "recce_state.json")
            shutil.copy(state_file, f.name)

            # load the state file
            state_loader = FileStateLoader(state_file=f.name)
            state = state_loader.load()
            assert len(state.runs) == 17


def test_lineage_diff(dbt_test_helper):
    sql_model1 = """
    select a from T
    """

    sql_model2 = """
    select a from {{ ref("model1") }}
    """

    sql_model2_ = """
    select
    a,b
    from
    {{ ref("model1") }}
    """

    dbt_test_helper.create_model("model1", sql_model1, sql_model1)
    dbt_test_helper.create_model("model2", sql_model2, sql_model2_)
    result = dbt_test_helper.context.get_lineage_diff()
    nodediff = result.diff.get("model1")
    assert nodediff is None
    nodediff2 = result.diff.get("model2")
    assert nodediff2 is not None and nodediff2.change_status == "modified"



from typing import Dict, Iterable, Set, Tuple

from recce.models.types import CllColumn, CllNode


def find_upstream(node_ids: Iterable, parent_map):
    visited = set()
    upstream = set()

    stack = list(node_ids)
    while stack:
        current = stack.pop()
        if current in visited:
            continue
        visited.add(current)
        parents = parent_map.get(current, [])
        for parent in parents:
            if parent not in upstream:
                upstream.add(parent)
                stack.append(parent)
    return upstream


def find_downstream(node_ids: Iterable, child_map):
    visited = set()
    downstream = set()

    stack = list(node_ids)
    while stack:
        current = stack.pop()
        if current in visited:
            continue
        visited.add(current)
        children = child_map.get(current, [])
        for child in children:
            if child not in downstream:
                downstream.add(child)
                stack.append(child)
    return downstream


def find_column_dependencies(node_column_id: str, parent_map: Dict, child_map: Dict) -> Tuple[Set, Set]:
    upstream_cols = find_upstream([node_column_id], parent_map)
    downstream_cols = find_downstream([node_column_id], child_map)
    return upstream_cols, downstream_cols


def filter_lineage_vertices(
    lineage_nodes: Dict[str, CllNode], lineage_columns: Dict[str, CllColumn], relevant_columns: Set[str]
) -> Tuple[Dict[str, CllNode], Dict[str, CllColumn]]:
    nodes = {}
    columns = {}

    for node_id, node in lineage_nodes.items():
        if node_id in relevant_columns:
            nodes[node_id] = node

    for col_id, column in lineage_columns.items():
        if col_id in relevant_columns:
            columns[col_id] = column

    return nodes, columns


def filter_dependency_maps(
    parent_map: Dict, child_map: Dict, relevant_ids: Set
) -> Tuple[Dict[str, Set], Dict[str, Set]]:
    p_map = {}
    c_map = {}
    for node_id, parents in parent_map.items():
        if node_id in relevant_ids:
            p_map[node_id] = {p for p in parents if p in relevant_ids}

    for node_id, children in child_map.items():
        if node_id in relevant_ids:
            c_map[node_id] = {c for c in children if c in relevant_ids}

    return p_map, c_map


def build_column_key(node_id: str, column_name: str) -> str:
    """Build a unique column key from node name and column name."""
    return f"{node_id}_{column_name}"



import json
import os
import platform
import sys
import time
from contextlib import contextmanager
from datetime import datetime, timezone
from json import JSONDecodeError

import portalocker
import requests

from recce import __version__, is_ci_env, is_recce_cloud_instance
from recce.github import is_github_codespace


class Collector:
    def __init__(self):
        self._api_endpoint = "https://api.amplitude.com/2/httpapi"
        self._api_key = None
        self._user_id = None

        self._unsend_events_file = None
        self._delete_threshold = 1000
        self._upload_threshold = 10
        self._is_ci: bool = is_ci_env()
        self._is_github_codespace: bool = is_github_codespace()
        self._is_recce_cloud_instance: bool = is_recce_cloud_instance()
        self._flush_timer = None

    def schedule_flush(self):
        def _callback():
            self._flush_timer = None
            self.send_events()

        # send async thread
        import threading

        if self._flush_timer:
            try:
                self._flush_timer.cancel()
            except Exception:
                # do nothing
                pass
        self._flush_timer = threading.Timer(1, _callback)
        self._flush_timer.start()

    def is_ready(self):
        if self._api_key is None or self._user_id is None:
            return False
        return True

    def set_api_key(self, api_key):
        self._api_key = api_key

    def set_user_id(self, user_id):
        self._user_id = user_id

    def set_unsend_events_file(self, unsend_events_file):
        self._unsend_events_file = unsend_events_file
        self._check_required_files()

    def _log_event(
        self,
        user_id,
        event_type,
        created_at,
        user_properties,
        event_properties,
    ):
        event = dict(
            user_id=user_id,
            event_type=event_type,
            ip="$remote",
            time=int(time.mktime(created_at.timetuple())),
            user_properties=user_properties,
            event_properties=event_properties,
            platform=sys.platform,
            os_version=platform.platform(),
            app_version=__version__,
        )
        self._store_to_file(event)
        if self._is_full():
            self.send_events()
        self._cleanup_unsend_events()

    def _get_user_id(self):
        # when the recce is running in automation use cases
        # replace the user id with project_id to avoid so many unique user id
        user_id = self._user_id
        if self._is_ci is True:
            user_id = f"{self._user_id}_CI"
        return user_id

    def log_event(self, event_prop, event_type, event_triggered_at: datetime = None, user_properties=None):
        # Use local timezone
        if event_triggered_at is None:
            created_at = datetime.now()
        else:
            # Convert to UTC timezone
            created_at = event_triggered_at.astimezone(timezone.utc)
        python_version = f"{sys.version_info.major}.{sys.version_info.minor}"

        # when the recce is running in automation use cases
        # replace the user id with project_id to avoid so many unique user id
        user_id = self._get_user_id()

        default_user_properties = dict(
            version=__version__,
            python_version=python_version,
            is_ci=self._is_ci,
            is_github_codespace=self._is_github_codespace,
            is_recce_cloud_instance=self._is_recce_cloud_instance,
        )

        if user_properties is not None:
            default_user_properties.update(user_properties)

        self._log_event(
            user_id=user_id,
            event_type=event_type,
            created_at=created_at,
            user_properties=default_user_properties,
            event_properties=event_prop,
        )

    def _check_required_files(self):
        user_home = os.path.dirname(self._unsend_events_file)
        if not os.path.exists(user_home):
            os.makedirs(user_home, exist_ok=True)
        if not os.path.exists(self._unsend_events_file):
            with portalocker.Lock(self._unsend_events_file, "w+", timeout=5) as f:
                f.write(json.dumps({"unsend_events": []}))

    def _is_full(self):
        with portalocker.Lock(self._unsend_events_file, "r+", timeout=5) as f:
            o = json.loads(f.read())
            return len(o.get("unsend_events", [])) >= self._upload_threshold

    @contextmanager
    def load_json(self):
        with portalocker.Lock(self._unsend_events_file, "r+", timeout=5) as f:
            o = None
            try:
                o = json.loads(f.read())
                yield o
            except JSONDecodeError:
                o = dict(unsend_events=[])
                yield o
            finally:
                f.seek(0)
                f.truncate()
                if o is not None:
                    f.write(json.dumps(o))

    def send_events(self):
        with self.load_json() as o:
            payload = dict(
                api_key=self._api_key,
                events=o["unsend_events"],
            )
            o["unsend_events"] = []
        try:
            requests.post(self._api_endpoint, json=payload)
        except Exception:
            # TODO: handle exception when sending events
            pass

    def _store_to_file(self, event):
        with self.load_json() as o:
            events = o.get("unsend_events", None)
            if events is None:
                o["unsend_events"] = []

            o["unsend_events"].append(event)

    def _cleanup_unsend_events(self):
        with self.load_json() as o:
            events = o.get("unsend_events", None)
            if events is None:
                o["unsend_events"] = []

            while len(o["unsend_events"]) > self._delete_threshold:
                o["unsend_events"].pop(0)



"""
Recce MCP (Model Context Protocol) Server

This module implements a stdio-based MCP server that provides tools for
interacting with Recce's data validation capabilities.
"""

import asyncio
import json
import logging
from typing import Any, Dict, List

from mcp.server import Server
from mcp.server.stdio import stdio_server
from mcp.types import TextContent, Tool

from recce.core import RecceContext, load_context
from recce.tasks.profile import ProfileDiffTask
from recce.tasks.query import QueryDiffTask, QueryTask
from recce.tasks.rowcount import RowCountDiffTask

logger = logging.getLogger(__name__)


class RecceMCPServer:
    """MCP Server for Recce data validation tools"""

    def __init__(self, context: RecceContext):
        self.context = context
        self.server = Server("recce")
        self._setup_handlers()

    def _setup_handlers(self):
        """Register all tool handlers"""

        @self.server.list_tools()
        async def list_tools() -> List[Tool]:
            """List all available tools"""
            return [
                Tool(
                    name="get_lineage_diff",
                    description="Get the lineage diff between base and current environments. "
                    "Shows added, removed, and modified models.",
                    inputSchema={
                        "type": "object",
                        "properties": {
                            "select": {
                                "type": "string",
                                "description": "dbt selector syntax to filter models (optional)",
                            },
                            "exclude": {
                                "type": "string",
                                "description": "dbt selector syntax to exclude models (optional)",
                            },
                            "packages": {
                                "type": "array",
                                "items": {"type": "string"},
                                "description": "List of packages to filter (optional)",
                            },
                        },
                    },
                ),
                Tool(
                    name="row_count_diff",
                    description="Compare row counts between base and current environments for specified models.",
                    inputSchema={
                        "type": "object",
                        "properties": {
                            "node_names": {
                                "type": "array",
                                "items": {"type": "string"},
                                "description": "List of model names to check row counts (optional)",
                            },
                            "node_ids": {
                                "type": "array",
                                "items": {"type": "string"},
                                "description": "List of node IDs to check row counts (optional)",
                            },
                            "select": {
                                "type": "string",
                                "description": "dbt selector syntax to filter models (optional)",
                            },
                            "exclude": {
                                "type": "string",
                                "description": "dbt selector syntax to exclude models (optional)",
                            },
                        },
                    },
                ),
                Tool(
                    name="query",
                    description="Execute a SQL query on the current environment. "
                    "Supports Jinja templates with dbt macros like {{ ref('model_name') }}.",
                    inputSchema={
                        "type": "object",
                        "properties": {
                            "sql_template": {
                                "type": "string",
                                "description": "SQL query template with optional Jinja syntax",
                            },
                            "base": {
                                "type": "boolean",
                                "description": "Whether to run on base environment (default: false)",
                                "default": False,
                            },
                        },
                        "required": ["sql_template"],
                    },
                ),
                Tool(
                    name="query_diff",
                    description="Execute SQL queries on both base and current environments and compare results. "
                    "Supports primary keys for row-level comparison.",
                    inputSchema={
                        "type": "object",
                        "properties": {
                            "sql_template": {
                                "type": "string",
                                "description": "SQL query template for current environment",
                            },
                            "base_sql_template": {
                                "type": "string",
                                "description": "SQL query template for base environment (optional, defaults to sql_template)",
                            },
                            "primary_keys": {
                                "type": "array",
                                "items": {"type": "string"},
                                "description": "List of primary key columns for row comparison (optional)",
                            },
                        },
                        "required": ["sql_template"],
                    },
                ),
                Tool(
                    name="profile_diff",
                    description="Generate and compare statistical profiles (min, max, avg, distinct count, etc.) "
                    "for columns in a model between base and current environments.",
                    inputSchema={
                        "type": "object",
                        "properties": {
                            "model": {
                                "type": "string",
                                "description": "Model name to profile",
                            },
                            "columns": {
                                "type": "array",
                                "items": {"type": "string"},
                                "description": "List of column names to profile (optional, profiles all columns if not specified)",
                            },
                        },
                        "required": ["model"],
                    },
                ),
            ]

        @self.server.call_tool()
        async def call_tool(name: str, arguments: Dict[str, Any]) -> List[TextContent]:
            """Handle tool calls"""
            try:
                if name == "get_lineage_diff":
                    result = await self._get_lineage_diff(arguments)
                elif name == "row_count_diff":
                    result = await self._row_count_diff(arguments)
                elif name == "query":
                    result = await self._query(arguments)
                elif name == "query_diff":
                    result = await self._query_diff(arguments)
                elif name == "profile_diff":
                    result = await self._profile_diff(arguments)
                else:
                    raise ValueError(f"Unknown tool: {name}")

                return [TextContent(type="text", text=json.dumps(result, indent=2))]
            except Exception as e:
                logger.exception(f"Error executing tool {name}")
                return [TextContent(type="text", text=json.dumps({"error": str(e)}, indent=2))]

    async def _get_lineage_diff(self, arguments: Dict[str, Any]) -> Dict[str, Any]:
        """Get lineage diff between base and current"""
        try:
            # Get lineage diff from adapter (returns a Pydantic LineageDiff model)
            lineage_diff = self.context.get_lineage_diff()
            # Convert Pydantic model to dict for JSON serialization
            return lineage_diff.model_dump(mode="json")
        except Exception:
            logger.exception("Error getting lineage diff")
            raise

    async def _row_count_diff(self, arguments: Dict[str, Any]) -> Dict[str, Any]:
        """Execute row count diff task"""
        try:
            task = RowCountDiffTask(params=arguments)

            # Execute task synchronously (it's already sync)
            result = await asyncio.get_event_loop().run_in_executor(None, task.execute)

            return result
        except Exception:
            logger.exception("Error executing row count diff")
            raise

    async def _query(self, arguments: Dict[str, Any]) -> Dict[str, Any]:
        """Execute a query"""
        try:
            sql_template = arguments.get("sql_template")
            is_base = arguments.get("base", False)

            params = {"sql_template": sql_template}
            task = QueryTask(params=params)
            task.is_base = is_base

            # Execute task
            result = await asyncio.get_event_loop().run_in_executor(None, task.execute)

            # Convert to dict if it's a model
            if hasattr(result, "model_dump"):
                return result.model_dump(mode="json")
            return result
        except Exception:
            logger.exception("Error executing query")
            raise

    async def _query_diff(self, arguments: Dict[str, Any]) -> Dict[str, Any]:
        """Execute query diff task"""
        try:
            task = QueryDiffTask(params=arguments)

            # Execute task
            result = await asyncio.get_event_loop().run_in_executor(None, task.execute)

            # Convert to dict if it's a model
            if hasattr(result, "model_dump"):
                return result.model_dump(mode="json")
            return result
        except Exception:
            logger.exception("Error executing query diff")
            raise

    async def _profile_diff(self, arguments: Dict[str, Any]) -> Dict[str, Any]:
        """Execute profile diff task"""
        try:
            task = ProfileDiffTask(params=arguments)

            # Execute task
            result = await asyncio.get_event_loop().run_in_executor(None, task.execute)

            # Convert to dict if it's a model
            if hasattr(result, "model_dump"):
                return result.model_dump(mode="json")
            return result
        except Exception:
            logger.exception("Error executing profile diff")
            raise

    async def run(self):
        """Run the MCP server"""
        async with stdio_server() as (read_stream, write_stream):
            await self.server.run(read_stream, write_stream, self.server.create_initialization_options())


async def run_mcp_server(**kwargs):
    """
    Entry point for running the MCP server

    Args:
        **kwargs: Arguments for loading RecceContext (dbt options, etc.)
    """
    # Setup logging
    logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s")

    # Load Recce context
    context = load_context(**kwargs)

    # Create and run server
    server = RecceMCPServer(context)
    await server.run()



import logging
import os
from typing import Optional, Tuple, Union

from .state import RecceState
from .state_loader import RecceStateLoader

logger = logging.getLogger("uvicorn")


class FileStateLoader(RecceStateLoader):
    def __init__(
        self,
        review_mode: bool = False,
        state_file: Optional[str] = None,
        initial_state: Optional[RecceState] = None,
    ):
        super().__init__(review_mode=review_mode, state_file=state_file, initial_state=initial_state)

    def verify(self) -> bool:
        if self.review_mode is True and self.state_file is None:
            self.error_message = "Recce can not launch without a state file."
            self.hint_message = "Please provide a state file in the command argument."
            return False
        return True

    def _load_state(self) -> Tuple[RecceState, str]:
        state = RecceState.from_file(self.state_file) if self.state_file else None
        state_tag = None
        return state, state_tag

    def _export_state(self, state: RecceState = None) -> Tuple[Union[str, None], str]:
        """
        Store the state to a file. Store happens when terminating the server or run instance.
        """

        if self.state_file is None:
            return "No state file is provided. Skip storing the state.", None

        logger.info(f"Store recce state to '{self.state_file}'")
        message = self._export_state_to_file(self.state_file)
        tag = None

        return message, tag

    def purge(self) -> bool:
        if self.state_file is not None:
            try:
                os.remove(self.state_file)
                return True
            except Exception as e:
                self.error_message = f"Failed to remove the state file: {e}"
                return False
        else:
            self.error_message = "No state file is provided. Skip removing the state file."
            return False



from dataclasses import dataclass

RECCE_STATE_FILE = "recce-state.json"
RECCE_STATE_COMPRESSED_FILE = f"{RECCE_STATE_FILE}.gz"


@dataclass
class ErrorMessage:
    error_message: str
    hint_message: str


RECCE_CLOUD_TOKEN_MISSING = ErrorMessage(
    error_message="No GitHub token is provided to access the pull request information",
    hint_message="Please provide a GitHub token in the command argument",
)

RECCE_CLOUD_PASSWORD_MISSING = ErrorMessage(
    error_message="No password provided to access the state file in Recce Cloud",
    hint_message='Please provide a password with the option "--password <compress-password>"',
)

RECCE_API_TOKEN_MISSING = ErrorMessage(
    error_message="No Recce API token is provided",
    hint_message="Please login to Recce Cloud and copy the API token from the settings page",
)



from typing import List, Optional

from recce.exceptions import RecceException

from .types import Check


class CheckDAO:
    """
    Data Access Object for Check. Currently, we store runs in memory, in the future, we can store them in a database.
    """

    @property
    def _checks(self):
        from recce.core import default_context

        return default_context().checks

    def create(self, check) -> None:
        self._checks.append(check)

    def find_check_by_id(self, check_id) -> Optional[Check]:
        for check in self._checks:
            if str(check_id) == str(check.check_id):
                return check

        return None

    def delete(self, check_id) -> bool:
        for check in self._checks:
            if str(check_id) == str(check.check_id):
                self._checks.remove(check)
                return True

        return False

    def list(self) -> List[Check]:
        return list(self._checks)

    def reorder(self, source: int, destination: int):

        if source < 0 or source >= len(self._checks):
            raise RecceException("Failed to reorder checks. Source index out of range")

        if destination < 0 or destination >= len(self._checks):
            raise RecceException("Failed to reorder checks. Destination index out of range")

        check_to_move = self._checks.pop(source)
        self._checks.insert(destination, check_to_move)

    def clear(self):
        self._checks.clear()

    def status(self):
        return {"total": len(self._checks), "approved": len([c for c in self._checks if c.is_checked])}



from unittest import TestCase
from unittest.mock import MagicMock, patch

from click.testing import CliRunner

from recce.cli import run as cli_command_run
from recce.cli import server as cli_command_server
from recce.cli import snapshot as cli_command_snapshot
from recce.cli import upload_session as cli_command_upload_session
from recce.core import RecceContext
from recce.state import CloudStateLoader


def test_cmd_version():
    from recce import __version__
    from recce.cli import version

    runner = CliRunner()
    result = runner.invoke(version, [])
    assert result.exit_code == 0
    assert result.output.replace("\n", "") == __version__


class TestCommandServer(TestCase):
    def setUp(self):
        self.runner = CliRunner()
        pass

    @patch.object(RecceContext, "verify_required_artifacts")
    @patch("recce.cli.uvicorn.run")
    def test_cmd_server(self, mock_run, mock_verify_required_artifacts):
        from recce.server import app

        mock_verify_required_artifacts.return_value = True, None
        self.runner.invoke(cli_command_server, ["--host", "unittest", "--port", 5566, "--single-env"])
        mock_run.assert_called_once_with(app, host="unittest", port=5566, lifespan="on")

    @patch("recce.cli.uvicorn.run")
    def test_cmd_server_with_cloud_without_password(self, mock_run):
        # Should fail if no password is provided
        result = self.runner.invoke(cli_command_server, ["--cloud"])
        assert result.exit_code == 1

    @patch("recce.cli.uvicorn.run")
    def test_cmd_server_with_cloud_without_token(self, mock_run):
        # Should fail if no token is provided
        result = self.runner.invoke(cli_command_server, ["--cloud", "--password", "unittest"])
        assert result.exit_code == 1

    @patch.object(RecceContext, "verify_required_artifacts")
    @patch("recce.util.recce_cloud.get_recce_cloud_onboarding_state")
    @patch("recce.cli.uvicorn.run")
    @patch("recce.cli.CloudStateLoader")
    def test_cmd_server_with_cloud(
        self, mock_state_loader_class, mock_run, mock_get_recce_cloud_onboarding_state, mock_verify_required_artifacts
    ):
        mock_state_loader = MagicMock(spec=CloudStateLoader)
        mock_state_loader.verify.return_value = True
        mock_state_loader.review_mode = True
        mock_get_recce_cloud_onboarding_state.return_value = "completed"
        mock_verify_required_artifacts.return_value = True, None

        mock_state_loader_class.return_value = mock_state_loader
        self.runner.invoke(
            cli_command_server, ["--cloud", "--password", "unittest", "--cloud-token", "unittest", "--single-env"]
        )
        mock_state_loader_class.assert_called_once()
        mock_run.assert_called_once()

    @patch.object(RecceContext, "verify_required_artifacts")
    @patch("recce.util.recce_cloud.get_recce_cloud_onboarding_state")
    @patch("recce.cli.uvicorn.run")
    @patch("recce.cli.CloudStateLoader")
    @patch("recce.cli.prepare_api_token", return_value="test_api_token")
    def test_cmd_server_with_session_id(
        self,
        mock_prepare_api_token,
        mock_state_loader_class,
        mock_run,
        mock_get_recce_cloud_onboarding_state,
        mock_verify_required_artifacts,
    ):
        """Test that --session-id automatically enables cloud and review mode"""
        mock_state_loader = MagicMock(spec=CloudStateLoader)
        mock_state_loader.verify.return_value = True
        mock_state_loader.review_mode = True
        mock_get_recce_cloud_onboarding_state.return_value = "completed"
        mock_verify_required_artifacts.return_value = True, None

        mock_state_loader_class.return_value = mock_state_loader

        # Test with --session-id (should automatically enable cloud and review)
        result = self.runner.invoke(cli_command_server, ["--session-id", "test-session-123", "--single-env"])

        # Should succeed
        assert result.exit_code == 0

        # Should create CloudStateLoader with session_id in cloud_options
        mock_state_loader_class.assert_called_once()
        call_args = mock_state_loader_class.call_args
        assert call_args.kwargs["review_mode"] is True
        assert "session_id" in call_args.kwargs["cloud_options"]
        assert call_args.kwargs["cloud_options"]["session_id"] == "test-session-123"

        mock_run.assert_called_once()

    @patch.object(RecceContext, "verify_required_artifacts")
    @patch("recce.util.recce_cloud.get_recce_cloud_onboarding_state")
    @patch("recce.cli.uvicorn.run")
    @patch("recce.cli.CloudStateLoader")
    @patch("recce.cli.prepare_api_token", return_value="test_api_token")
    def test_cmd_server_with_share_url(
        self,
        mock_prepare_api_token,
        mock_state_loader_class,
        mock_run,
        mock_get_recce_cloud_onboarding_state,
        mock_verify_required_artifacts,
    ):
        """Test that --share-url automatically enables cloud and review mode"""
        mock_state_loader = MagicMock(spec=CloudStateLoader)
        mock_state_loader.verify.return_value = True
        mock_state_loader.review_mode = True
        mock_get_recce_cloud_onboarding_state.return_value = "completed"
        mock_verify_required_artifacts.return_value = True, None

        mock_state_loader_class.return_value = mock_state_loader

        # Test with --share-url (should automatically enable cloud and review)
        result = self.runner.invoke(
            cli_command_server, ["--share-url", "https://cloud.recce.io/share/abc123", "--single-env"]
        )

        # Should succeed
        assert result.exit_code == 0

        # Should create CloudStateLoader with share_id in cloud_options
        mock_state_loader_class.assert_called_once()
        call_args = mock_state_loader_class.call_args
        assert call_args.kwargs["review_mode"] is True
        assert "share_id" in call_args.kwargs["cloud_options"]
        assert call_args.kwargs["cloud_options"]["share_id"] == "abc123"

        mock_run.assert_called_once()

    @patch.object(RecceContext, "verify_required_artifacts")
    @patch("os.path.isdir", side_effect=lambda path: True if path == "existed_folder" else False)
    @patch("recce.cli.uvicorn.run")
    @patch("recce.server.AppState")
    def test_cmd_server_with_single_env(self, mock_app_state, mock_run, mock_isdir, mock_verify_required_artifacts):
        mock_verify_required_artifacts.return_value = True, None
        self.runner.invoke(
            cli_command_server,
            ["--single-env", "--target-path", "existed_folder", "--target-base-path", "non_existed_folder"],
        )
        mock_run.assert_called_once()

        # Onboarding mode should be set to True
        app_state_call_args = mock_app_state.call_args
        app_state_flag = app_state_call_args.kwargs["flag"]
        assert "single_env_onboarding" in app_state_flag
        assert app_state_flag["single_env_onboarding"] is True
        assert "show_relaunch_hint" in app_state_flag
        assert app_state_flag["show_relaunch_hint"] is True

        # The target_base_path should be set to the same as target_path
        verify_required_artifacts_args = mock_verify_required_artifacts.call_args
        assert (
            verify_required_artifacts_args.kwargs["target_path"]
            == verify_required_artifacts_args.kwargs["target_base_path"]
        )

    @patch.object(RecceContext, "verify_required_artifacts")
    @patch("os.path.isdir", side_effect=lambda path: True if path == "existed_folder" else False)
    @patch("recce.cli.uvicorn.run")
    @patch("recce.server.AppState")
    def test_cmd_server_with_single_env_but_review_mode_enabled(
        self, mock_app_state, mock_run, mock_isdir, mock_verify_required_artifacts
    ):
        mock_verify_required_artifacts.return_value = True, None
        self.runner.invoke(
            cli_command_server,
            [
                "existed_state_file",
                "--review",
                "--single-env",
                "--target-path",
                "existed_folder",
                "--target-base-path",
                "non_existed_folder",
            ],
        )
        mock_run.assert_called_once()
        app_state_call_args = mock_app_state.call_args
        app_state_flag = app_state_call_args.kwargs["flag"]
        assert "single_env_onboarding" in app_state_flag
        assert app_state_flag["single_env_onboarding"] is False


class TestCommandRun(TestCase):
    def setUp(self):
        self.runner = CliRunner()
        pass

    @patch.object(RecceContext, "verify_required_artifacts")
    @patch("recce.cli.cli_run")
    def test_cmd_run(self, mock_cli_run, mock_verify_required_artifacts):
        mock_verify_required_artifacts.return_value = True, None

        self.runner.invoke(cli_command_run, [])
        mock_cli_run.assert_called_once()


class TestCommandUploadSession(TestCase):
    def setUp(self):
        self.runner = CliRunner()
        pass

    @patch("recce.cli.prepare_api_token", return_value="unittest_token")
    @patch("recce.cli.upload_artifacts_to_session", return_value=0)
    def test_cmd_upload_session(self, mock_upload_artifacts_to_session, mock_prepare_api_token):
        self.runner.invoke(
            cli_command_upload_session,
            ["--session-id", "unittest_session", "--api-token", mock_prepare_api_token.return_value],
        )
        mock_upload_artifacts_to_session.assert_called_once_with(
            "target", session_id="unittest_session", token="unittest_token", debug=False
        )

        self.runner.invoke(
            cli_command_snapshot,
            ["--snapshot-id", "unittest_session", "--api-token", mock_prepare_api_token.return_value],
        )
        mock_upload_artifacts_to_session.assert_called_once_with(
            "target", session_id="unittest_session", token="unittest_token", debug=False
        )



import os

import requests
from packaging.version import Version


def is_ci_env():
    # List of CI environment variables and their expected values
    ci_environments = {
        "CI": "true",  # Generic CI indicator
        "CIRCLECI": "true",  # CircleCI
        "GITHUB_ACTIONS": "true",  # GitHub Actions
        "GITLAB_CI": "true",  # GitLab CI
        "JENKINS_URL": None,  # Jenkins (just needs to exist)
        "TRAVIS": "true",  # Travis CI
        "APPVEYOR": "true",  # AppVeyor
        "DRONE": "true",  # Drone CI
        "TEAMCITY_VERSION": None,  # TeamCity
        "BITBUCKET_COMMIT": None,  # Bitbucket Pipelines
        "BUILDKITE": "true",  # Buildkite
        "CODEBUILD_BUILD_ID": None,  # AWS CodeBuild
        "AZURE_PIPELINES": "true",  # Azure Pipelines
    }

    for env_var, expected_value in ci_environments.items():
        env_value = os.environ.get(env_var)
        if env_value is not None:
            # If we just need the variable to exist
            if expected_value is None:
                return True
            # If we need to match a specific value (case-insensitive)
            if env_value.lower() == expected_value.lower():
                return True

    return False


def is_recce_cloud_instance():
    """Check if running in Recce Cloud instance."""
    return os.environ.get("RECCE_CLOUD_INSTANCE", "false").lower() == "true"


def get_runner():
    # GitHub Action
    if os.environ.get("GITHUB_ACTIONS", "false") == "true":
        return "github actions"

    # GitHub Codespace
    if os.environ.get("CODESPACES", "false") == "true":
        return "github codespaces"

    # CircleCI
    if os.environ.get("CIRCLECI", "false") == "true":
        return "circleci"

    return None


def get_version():
    version_file = os.path.normpath(os.path.join(os.path.dirname(__file__), "VERSION"))
    with open(version_file) as fh:
        version = fh.read().strip()
        return version


def fetch_latest_version():
    current_version = get_version()
    if "dev" in current_version:
        # Skip fetching latest version if it's a dev version
        return current_version

    try:
        url = "https://pypi.org/pypi/recce/json"
        response = requests.get(url, timeout=3)
        response.raise_for_status()
        data = response.json()
        return data["info"]["version"]
    except Exception:
        return current_version


__version__ = get_version()
__latest_version__ = fetch_latest_version()
__is_recce_outdated__ = Version(__version__) < Version(__latest_version__)



import gzip
import json
import os
import shutil
import tarfile
import tempfile
from urllib.parse import urlencode

import requests
from rich.console import Console

from recce.git import commit_hash_from_branch, current_branch, hosting_repo
from recce.state import s3_sse_c_headers
from recce.util.recce_cloud import PresignedUrlMethod, RecceCloud


def verify_artifacts_path(target_path: str) -> bool:
    """
    Verify if the target path is a valid artifacts path.

    :param target_path: the target path to check
    :return: True if the target path is a valid artifacts path, False otherwise
    """
    if not target_path:
        return False

    if not os.path.exists(target_path):
        return False

    if not os.path.isdir(target_path):
        return False

    required_artifacts_files = ["manifest.json", "catalog.json"]

    if all(f in os.listdir(target_path) for f in required_artifacts_files):
        # Check if the required files are present in the target path
        return True

    return False


def parse_dbt_version(file_path: str) -> str:
    with open(file_path, "r", encoding="utf-8") as f:
        data = json.load(f)

    dbt_version = data.get("metadata", {}).get("dbt_version", None)
    return dbt_version


def archive_artifacts(target_path: str) -> (str, str):
    if verify_artifacts_path(target_path) is False:
        raise Exception(f"Invalid target path: {target_path}")

    manifest_path = os.path.join(target_path, "manifest.json")
    catalog_path = os.path.join(target_path, "catalog.json")

    dbt_version = parse_dbt_version(manifest_path)
    if dbt_version is None:
        raise Exception("Failed to parse dbt version from manifest.json")

    # prepare the temporary artifacts path
    tmp_dir = tempfile.mkdtemp()
    artifacts_tar_path = os.path.join(tmp_dir, "dbt_artifacts.tar")
    artifacts_tar_gz_path = artifacts_tar_path + ".gz"

    with tarfile.open(artifacts_tar_path, "w") as tar:
        tar.add(manifest_path, arcname="manifest.json")
        tar.add(catalog_path, arcname="catalog.json")

    # Compress the tar file
    with open(artifacts_tar_path, "rb") as f_in, gzip.open(artifacts_tar_gz_path, "wb") as f_out:
        f_out.writelines(f_in)

    # Clean up the temporary directory
    try:
        os.remove(artifacts_tar_path)
    except FileNotFoundError:
        pass

    return artifacts_tar_gz_path, dbt_version


def upload_artifacts_to_session(target_path: str, session_id: str, token: str, debug: bool = False):
    """Upload dbt artifacts to a specific session ID in Recce Cloud."""
    console = Console()
    if verify_artifacts_path(target_path) is False:
        console.print(f"[[red]Error[/red]] Invalid target path: {target_path}")
        console.print("Please provide a valid target path containing manifest.json and catalog.json.")
        return 1

    manifest_path = os.path.join(target_path, "manifest.json")
    catalog_path = os.path.join(target_path, "catalog.json")

    # get the adapter type from the manifest file
    with open(manifest_path, "r", encoding="utf-8") as f:
        manifest_data = json.load(f)
        adapter_type = manifest_data.get("metadata", {}).get("adapter_type")
        if adapter_type is None:
            raise Exception("Failed to parse adapter type from manifest.json")

    recce_cloud = RecceCloud(token)

    session = recce_cloud.get_session(session_id)

    org_id = session.get("org_id")
    if org_id is None:
        raise Exception(f"Session ID {session_id} does not belong to any organization.")

    project_id = session.get("project_id")
    if project_id is None:
        raise Exception(f"Session ID {session_id} does not belong to any project.")

    # Get the presigned URL for uploading the artifacts using session ID
    console.print(f'Uploading artifacts for session ID "{session_id}"')
    presigned_urls = recce_cloud.get_upload_urls_by_session_id(org_id, project_id, session_id)
    if debug:
        console.rule("Debug information", style="blue")
        console.print(f"Org ID: {org_id}")
        console.print(f"Project ID: {project_id}")
        console.print(f"Session ID: {session_id}")
        console.print(f"Manifest path: {presigned_urls['manifest_url']}")
        console.print(f"Catalog path: {presigned_urls['catalog_url']}")
        console.print(f"Adapter type: {adapter_type}")

    # Upload the compressed artifacts (no password needed for session uploads)
    console.print(f'Uploading manifest from path "{manifest_path}"')
    response = requests.put(presigned_urls["manifest_url"], data=open(manifest_path, "rb").read())
    if response.status_code != 200 and response.status_code != 204:
        raise Exception(response.text)
    console.print(f'Uploading catalog from path "{catalog_path}"')
    response = requests.put(presigned_urls["catalog_url"], data=open(catalog_path, "rb").read())
    if response.status_code != 200 and response.status_code != 204:
        raise Exception(response.text)

    # Update the session metadata
    recce_cloud.update_session(org_id, project_id, session_id, adapter_type)

    return 0


def upload_dbt_artifacts(target_path: str, branch: str, token: str, password: str, debug: bool = False):
    console = Console()
    if verify_artifacts_path(target_path) is False:
        console.print(f"[[red]Error[/red]] Invalid target path: {target_path}")
        console.print("Please provide a valid target path containing manifest.json and catalog.json.")
        return 1

    if branch != current_branch():
        console.print(
            f"[[yellow]Warning[/yellow]] You are uploading the dbt artifacts as branch '{branch}'. "
            f"However, the current branch is '{current_branch()}'."
        )
        console.print("Please make sure you are uploading the dbt artifacts to the correct branch.")

    compress_file_path, dbt_version = archive_artifacts(target_path)
    repo = hosting_repo()
    sha = commit_hash_from_branch(branch)
    metadata = {"commit": sha, "dbt_version": dbt_version}

    # Get the presigned URL for uploading the artifacts
    presigned_url = RecceCloud(token).get_presigned_url_by_github_repo(
        method=PresignedUrlMethod.UPLOAD,
        repository=repo,
        artifact_name="dbt_artifacts.tar.gz",
        branch=branch,
        metadata=metadata,
    )

    if debug:
        console.rule("Debug information", style="blue")
        console.print(f"Branch: {branch}")
        console.print(f"Commit hash: {sha}")
        console.print(f"GitHub repository: {repo}")
        console.print(f"Artifact path: {compress_file_path}")
        console.print(f"DBT version: {dbt_version}")
        console.print(f"Presigned URL: {presigned_url}")
    console.print(f'Uploading the dbt artifacts from path "{target_path}" to branch "{branch}"')

    # Upload the compressed artifacts

    headers = s3_sse_c_headers(password)
    if metadata:
        headers["x-amz-tagging"] = urlencode(metadata)
    response = requests.put(presigned_url, data=open(compress_file_path, "rb").read(), headers=headers)
    if response.status_code != 200:
        raise Exception({response.text})

    # Clean up the compressed artifacts
    try:
        # Remove the compressed artifacts
        os.remove(compress_file_path)
        # Clean up the temporary directory
        os.rmdir(os.path.dirname(compress_file_path))
    except FileNotFoundError:
        pass


def download_dbt_artifacts(
    target_path: str, branch: str, token: str, password: str, force: bool = False, debug: bool = False
):
    console = Console()
    repo = hosting_repo()
    sha = None
    dbt_version = None

    presigned_url, tags = RecceCloud(token).get_download_presigned_url_by_github_repo_with_tags(
        repository=repo,
        artifact_name="dbt_artifacts.tar.gz",
        branch=branch,
    )
    if tags:
        sha = tags.get("commit")
        dbt_version = tags.get("dbt_version")

    if debug:
        console.rule("Debug information", style="blue")
        console.print(f"Git Branch: {branch}")
        console.print(f"Git Commit: {sha}")
        console.print(f"GitHub repository: {repo}")
        console.print(f"DBT version: {dbt_version}")
    console.print(f'Downloading from branch: "{branch}" and extracting to "{target_path}"')

    headers = s3_sse_c_headers(password)
    response = requests.get(presigned_url, headers=headers)

    if response.status_code != 200:
        raise Exception(response.text)

    if os.path.exists(target_path):
        if not force:
            raise Exception(
                f"Path {target_path} already exists. Please provide a new path or use '--force' option to overwrite the existing folder."
            )
        console.print(f"[[yellow]Warning[/yellow]] Overwrite existing path: {target_path}")
        shutil.rmtree(target_path)
    os.mkdir(target_path)

    tar_gz_file = os.path.join(target_path, "dbt_artifacts.tar.gz")
    with open(tar_gz_file, "wb") as f:
        f.write(response.content)

    with tarfile.open(tar_gz_file, "r") as tar:
        tar.extractall(path=target_path)

    # Clean up the compressed artifacts
    try:
        # Remove the compressed artifacts
        os.remove(tar_gz_file)
    except FileNotFoundError:
        pass
    return 0


def delete_dbt_artifacts(branch: str, token: str, debug: bool = False):
    """Delete dbt artifacts from a specific branch in Recce Cloud."""
    console = Console()
    repo = hosting_repo()

    if debug:
        console.rule("Debug information", style="blue")
        console.print(f"Git Branch: {branch}")
        console.print(f"GitHub repository: {repo}")

    console.print(f'Deleting dbt artifacts from branch: "{branch}"')

    RecceCloud(token).purge_artifacts(repo, branch=branch)



import pytest

from recce.tasks import RowCountDiffTask


def test_row_count(dbt_test_helper):
    csv_data_curr = """
        customer_id,name,age
        1,Alice,30
        2,Bob,25
        3,Charlie,35
        """

    csv_data_base = """
        customer_id,name,age
        1,Alice,35
        2,Bob,25
        """

    dbt_test_helper.create_model("customers", csv_data_base, csv_data_curr, unique_id="model.customers")
    task = RowCountDiffTask(dict(node_names=["customers"]))
    run_result = task.execute()
    assert run_result["customers"]["base"] == 2
    assert run_result["customers"]["curr"] == 3

    task = RowCountDiffTask(dict(node_names=["customers_"]))
    run_result = task.execute()
    assert run_result["customers_"]["base"] is None
    assert run_result["customers_"]["curr"] is None

    task = RowCountDiffTask(dict(node_ids=["model.customers"]))
    run_result = task.execute()
    assert run_result["customers"]["base"] == 2
    assert run_result["customers"]["curr"] == 3


def test_row_count_with_selector(dbt_test_helper):
    csv_data_1 = """
        customer_id,name,age
        1,Alice,30
        2,Bob,25
        3,Charlie,35
        """

    csv_data_2 = """
        customer_id,name,age
        1,Alice,35
        2,Bob,25
        """

    dbt_test_helper.create_model("model_1", csv_data_1, csv_data_2, depends_on=[])
    dbt_test_helper.create_model(
        "model_2", csv_data_1, csv_data_1, depends_on=["model_1"], package_name="other_package"
    )
    task = RowCountDiffTask(dict(select="model_1"))
    run_result = task.execute()
    assert len(run_result) == 1

    task = RowCountDiffTask(dict(select="model_1+"))
    run_result = task.execute()
    assert len(run_result) == 2


def test_validator():
    from recce.tasks.rowcount import RowCountDiffCheckValidator

    validator = RowCountDiffCheckValidator()

    def validate(params: dict):
        validator.validate(
            {
                "name": "test",
                "type": "row_count_diff",
                "params": params,
            }
        )

    # Select all modesl
    validate({})

    # Select by node name
    validate(
        {
            "node_names": ["abc"],
        }
    )
    with pytest.raises(ValueError):
        validate(
            {
                "node_names": "abc",
            }
        )

    # Select by node id
    validate(
        {
            "node_ids": ["model.abc"],
        }
    )

    # Select by selector
    validate(
        {
            "select": "customers",
            "exclude": "customers",
            "packages": ["jaffle_shop"],
            "view_mode": "all",
        }
    )

    # packages should be an array
    with pytest.raises(ValueError):
        validate(
            {
                "packages": "jaffle_shop",
            }
        )

    # view_mode should be 'all' or 'changed_models'
    validate(
        {
            "view_mode": None,
        }
    )
    validate(
        {
            "view_mode": "all",
        }
    )
    with pytest.raises(ValueError):
        validate(
            {
                "view_mode": "abc",
            }
        )



from typing import List, Optional, TypedDict, Union

from pydantic import BaseModel

from ..core import default_context
from ..exceptions import RecceException
from ..models import Check
from .core import CheckValidator, Task, TaskResultDiffer
from .dataframe import DataFrame


class ValueDiffParams(BaseModel):
    model: str
    primary_key: Union[str, List[str]]
    columns: Optional[List[str]] = None


class ValueDiffResult(BaseModel):
    class Summary(BaseModel):
        total: int
        added: int
        removed: int

    summary: Summary
    data: DataFrame


class ValueDiffMixin:
    def _verify_primary_key(self, dbt_adapter, primary_key: Union[str, List[str]], model: str):
        self.update_progress(message=f"Verify primary key: {primary_key}")
        composite = True if isinstance(primary_key, List) else False

        if composite:
            if len(primary_key) == 0:
                raise RecceException("Primary key cannot be empty")
            sql_template = r"""
            {%- set column_list = primary_key %}
            {%- set columns_csv = column_list | join(', ') %}

            with validation_errors as (
                select
                    {{ columns_csv }}
                from {{ relation }}
                group by {{ columns_csv }}
                having count(*) > 1
            )

            select *
            from validation_errors
            """
        else:
            if primary_key is None or len(primary_key) == 0:
                raise RecceException("Primary key cannot be empty")
            sql_template = r"""{{ adapter.dispatch('test_unique', 'dbt')(relation, primary_key) }}"""

        # check primary keys
        for base in [True, False]:
            relation = dbt_adapter.create_relation(model, base)
            context = dict(
                relation=relation,
                primary_key=primary_key,
            )

            sql = dbt_adapter.generate_sql(sql_template, context=context)
            sql_test = f"""SELECT COUNT(*) AS INVALIDS FROM ({sql}) AS T"""

            response, table = dbt_adapter.adapter.execute(sql_test, fetch=True)
            for row in table.rows:
                invalids = row[0]
                if invalids > 0:
                    raise RecceException(
                        f"Invalid primary key: \"{primary_key}\". The column should be unique. Please check by this sql: '{sql}'"
                    )
                break
            else:
                # it will never happen unless we use a wrong check sql
                raise RecceException("Cannot verify primary key")


class ValueDiffTask(Task, ValueDiffMixin):
    def __init__(self, params):
        super().__init__()
        self.params = ValueDiffParams(**params)
        self.connection = None
        self.legacy_surrogate_key = True

    def _query_value_diff(
        self,
        dbt_adapter,
        primary_key: Union[str, List[str]],
        model: str,
        columns: List[str] = None,
    ):
        import agate

        column_groups = {}
        composite = True if isinstance(primary_key, List) else False

        if columns is None or len(columns) == 0:
            base_columns = [column.column for column in dbt_adapter.get_columns(model, base=True)]
            curr_columns = [column.column for column in dbt_adapter.get_columns(model, base=False)]
            columns = [column for column in base_columns if column in curr_columns]
        completed = 0

        if composite:
            for primary_key_comp in primary_key[::-1]:
                if primary_key_comp not in columns:
                    columns.insert(0, primary_key_comp)
        else:
            if primary_key not in columns:
                columns.insert(0, primary_key)

        sql_template = r"""
        {%- set default_null_value = "_recce_surrogate_key_null_" -%}
        {%- set fields = [] -%}

        {%- for field in primary_keys -%}
            {%- do fields.append(
                "coalesce(cast(" ~ field ~ " as " ~ dbt.type_string() ~ "), '" ~ default_null_value  ~"')"
            ) -%}

            {%- if not loop.last %}
                {%- do fields.append("'-'") -%}
            {%- endif -%}
        {%- endfor -%}

        {%- set _pk = dbt.hash(dbt.concat(fields)) -%}

        with a_query as (
            select {{ _pk }} as _pk, * from {{ base_relation }}
        ),

        b_query as (
            select {{ _pk }} as _pk, * from {{ curr_relation }}
        ),

        joined as (
            select
                coalesce(a_query._pk, b_query._pk) as _pk,
                a_query.{{ column_to_compare }} as a_query_value,
                b_query.{{ column_to_compare }} as b_query_value,
                case
                    when a_query.{{ column_to_compare }} = b_query.{{ column_to_compare }} then 'perfect match'
                    when a_query.{{ column_to_compare }} is null and b_query.{{ column_to_compare }} is null then 'both are null'
                    when a_query._pk is null then 'missing from {{ a_relation_name }}'
                    when b_query._pk is null then 'missing from {{ b_relation_name }}'
                    when a_query.{{ column_to_compare }} is null then 'value is null in {{ a_relation_name }} only'
                    when b_query.{{ column_to_compare }} is null then 'value is null in {{ b_relation_name }} only'
                    when a_query.{{ column_to_compare }} != b_query.{{ column_to_compare }} then 'values do not match'
                    else 'unknown' -- this should never happen
                end as match_status
            from a_query
            full outer join b_query on a_query._pk = b_query._pk
        ),

        aggregated as (
            select
                '{{ column_to_compare }}' as column_name,
                match_status,
                count(*) as count_records
            from joined
            group by 1, 2
        )

        select
            column_name,
            match_status,
            count_records,
            round(100.0 * count_records / sum(count_records) over (), 2) as percent_of_total
        from aggregated
        """

        for column in columns:
            self.update_progress(message=f"Diff column: {column}", percentage=completed / len(columns))

            sql = dbt_adapter.generate_sql(
                sql_template,
                context=dict(
                    base_relation=dbt_adapter.create_relation(model, base=True),
                    curr_relation=dbt_adapter.create_relation(model, base=False),
                    primary_keys=primary_key if composite else [primary_key],
                    column_to_compare=column,
                ),
            )

            _, table = dbt_adapter.execute(sql, fetch=True)
            if column not in column_groups:
                column_groups[column] = dict(added=0, removed=0, mismatched=0, matched=0)
            for row in table.rows:
                # data example:
                # ('COLUMN_NAME', 'MATCH_STATUS', 'COUNT_RECORDS', 'PERCENT_OF_TOTAL')
                # ('EVENT_ID', 'perfect match', 158601510, Decimal('100.00'))
                column_name, column_state, row_count, total_rate = row
                if "column_name" == row[0].lower():
                    # skip column names
                    return

                # sample data like this:
                #     https://github.com/dbt-labs/dbt-audit-helper/blob/main/macros/compare_column_values.sql
                #
                #     'perfect match'            -> matched
                #     'both are null'            -> matched
                #     'missing from a'           -> row added
                #     'missing from b'           -> row removed
                #     'value is null in a only'  -> mismatched
                #     'value is null in b only'  -> mismatched
                #     'values do not match'      -> mismatched
                #     'unknown'                  -> this should never happen
                # end as match_status,

                state_mappings = {
                    "perfect match": "matched",
                    "both are null": "matched",
                    "missing from a": "added",
                    "missing from b": "removed",
                    "value is null in a only": "mismatched",
                    "value is null in b only": "mismatched",
                    "values do not match": "mismatched",
                }

                # Use the mapping to update counts
                for state, action in state_mappings.items():
                    if state in column_state:
                        column_groups[column_name][action] += row_count

            # Cancel as early as possible
            self.check_cancel()

            completed = completed + 1

        first = list(column_groups.values())[0]
        added = first["added"]
        removed = first["removed"]
        common = first["matched"] + first["mismatched"]
        total = common + added + removed

        row = []
        for k, v in column_groups.items():
            if composite and k.lower() == "_pk":
                continue
            # This is incorrect when there are one side null
            # https://github.com/dbt-labs/dbt-audit-helper/blob/main/macros/compare_column_values.sql#L20-L23
            # matched = v['matched']
            matched = common - v["mismatched"]
            rate = None if common == 0 else matched / common
            record = [k, matched, rate]
            row.append(record)

        column_names = ["column", "matched", "matched_p"]
        column_types = [agate.Text(), agate.Number(), agate.Number()]
        table = agate.Table(row, column_names=column_names, column_types=column_types)

        return ValueDiffResult(
            summary=ValueDiffResult.Summary(total=total, added=added, removed=removed),
            data=DataFrame.from_agate(table),
        )

    def execute(self):
        dbt_adapter = default_context().adapter

        with dbt_adapter.connection_named("value diff"):
            self.connection = dbt_adapter.get_thread_connection()

            primary_key: Union[str, List[str]] = self.params.primary_key
            model: str = self.params.model
            columns: List[str] = self.params.columns

            self._verify_primary_key(dbt_adapter, primary_key, model)
            self.check_cancel()

            return self._query_value_diff(dbt_adapter, primary_key, model, columns=columns)

    def cancel(self):
        super().cancel()
        from recce.adapter.dbt_adapter import DbtAdapter

        if self.connection:
            adapter: DbtAdapter = default_context().adapter
            with adapter.connection_named("cancel"):
                adapter.cancel(self.connection)


class ValueDiffTaskResultDiffer(TaskResultDiffer):
    def _check_result_changed_fn(self, result):
        is_changed = False
        summary = result.get("summary", {})
        added = summary.get("added", 0)
        removed = summary.get("removed", 0)
        changes = {"column_changed": []}

        if added > 0:
            is_changed = True
            changes["row_added"] = added

        if removed > 0:
            is_changed = True
            changes["row_removed"] = removed

        row_data = result.get("data", {}).get("data", [])
        for row in row_data:
            column, matched, matched_p = row
            if float(matched_p) < 1.0:
                # if there is any mismatched, we consider it as changed
                is_changed = True
                changes["column_changed"].append(
                    {
                        "column": column,
                        "matched": matched,
                        "matched_p": matched_p,
                    }
                )

        return changes if is_changed else None


class ValueDiffDetailParams(TypedDict):
    primary_key: str
    model: str
    columns: List[str]


class ValueDiffDetailResult(DataFrame):
    pass


class ValueDiffDetailTask(Task, ValueDiffMixin):
    def __init__(self, params):
        super().__init__()
        self.params = ValueDiffParams(**params)
        self.connection = None
        self.legacy_surrogate_key = True

    def _query_value_diff(
        self,
        dbt_adapter,
        primary_key: Union[str, List[str]],
        model: str,
        columns: List[str] = None,
    ):
        composite = True if isinstance(primary_key, List) else False

        if columns is None or len(columns) == 0:
            base_columns = [column.column for column in dbt_adapter.get_columns(model, base=True)]
            curr_columns = [column.column for column in dbt_adapter.get_columns(model, base=False)]
            columns = [column for column in base_columns if column in curr_columns]

        if composite:
            for primary_key_comp in primary_key[::-1]:
                if primary_key_comp not in columns:
                    columns.insert(0, primary_key_comp)
        else:
            if primary_key not in columns:
                columns.insert(0, primary_key)

        sql_template = r"""
        with a_query as (
            select {{ columns | join(',\n') }} from {{ base_relation }}
        ),

        b_query as (
            select {{ columns | join(',\n') }} from {{ curr_relation }}
        ),

        a_intersect_b as (
            select * from a_query
            {{ dbt.intersect() }}
            select * from b_query
        ),

        a_except_b as (
            select * from a_query
            {{ dbt.except() }}
            select * from b_query
        ),

        b_except_a as (
            select * from b_query
            {{ dbt.except() }}
            select * from a_query
        ),

        all_records as (
            select
                *,
                true as in_a,
                true as in_b
            from a_intersect_b

            union all

            select
                *,
                true as in_a,
                false as in_b
            from a_except_b

            union all

            select
                *,
                false as in_a,
                true as in_b
            from b_except_a
        )

        select * from all_records
        where not (in_a and in_b)
        order by {{ primary_keys | join(',\n') }}, in_a desc, in_b desc
        limit {{ limit }}
        """

        sql = dbt_adapter.generate_sql(
            sql_template,
            context=dict(
                base_relation=dbt_adapter.create_relation(model, base=True),
                curr_relation=dbt_adapter.create_relation(model, base=False),
                primary_keys=primary_key if composite else [primary_key],
                columns=columns,
                limit=1000,
            ),
        )

        _, table = dbt_adapter.execute(sql, fetch=True)
        self.check_cancel()

        return DataFrame.from_agate(table)

    def execute(self):
        from recce.adapter.dbt_adapter import DbtAdapter

        dbt_adapter: DbtAdapter = default_context().adapter

        with dbt_adapter.connection_named("value diff"):
            self.connection = dbt_adapter.get_thread_connection()

            primary_key: Union[str, List[str]] = self.params.primary_key
            model: str = self.params.model
            columns: List[str] = self.params.columns

            self._verify_primary_key(dbt_adapter, primary_key, model)
            self.check_cancel()

            return self._query_value_diff(dbt_adapter, primary_key, model, columns)

    def cancel(self):
        from recce.adapter.dbt_adapter import DbtAdapter

        if self.connection:
            adapter: DbtAdapter = default_context().adapter
            with adapter.connection_named("cancel"):
                adapter.cancel(self.connection)


class ValueDiffDetailTaskResultDiffer(TaskResultDiffer):
    def _check_result_changed_fn(self, result):
        diff_data = result.get("data")
        if diff_data is None or len(diff_data) == 0:
            return None

        # TODO: Implement detailed information of values changed
        return dict(values_changed={})


class ValueDiffCheckValidator(CheckValidator):
    def validate_check(self, check: Check):
        try:
            ValueDiffParams(**check.params)
        except Exception as e:
            raise ValueError(f"Invalid check: {str(e)}")
